# ./coc-enter.vim

# print("__package__: " + __package__)

if __package__ is None:
    from os.path import sep
    *_, __package__, _ = __file__.split(sep)

from .timer import Timer   # Single Use
timer = Timer()
timer.start("test.py")
from .color import Code    # Multi-Use
from .error import handler # Single Use
from .ic import ic         # Multi-Use
import time

def function_call(item_1, key=[1, 2, 3], D={}):
    pass

def print_int(i):
    ic(Code.CYAN + i)

def print_list(L):
    five = 5
    zero = 0
    infinity = five / zero
    ic(Code.MAGENTA + L)

def print_str(s):
    ic(Code.GREEN + s)

def func(arg):
    if type(arg) == int:
        print_int(arg)
    elif type(arg) == list:
        print_list(arg)
    elif type(arg) == str:
        print_str(arg)

def caller():
    import sys
    import inspect
    stack = inspect.stack()
    frame2 = sys._getframe(1)



def git_is_dirty():
    import git
    path = r"C:\Users\vivek\Desktop\bkp\game"
    repo = git.Repo(path)
    index = False
    working_tree = False
    untracked_files = False
    submodules = False
    timer.start("is_dirty")
    ic(index, working_tree, untracked_files, submodules)
    breakpoint()
    is_dirty = repo.is_dirty(index=index, working_tree=working_tree, untracked_files=untracked_files, submodules=submodules, path='file-022.txt')
    ic(is_dirty)
    timer.print("is_dirty")
    time.sleep(1)

def sqlite_init():
    import os.path
    from .db import db_init
    temp_directory = os.path.expandvars("%TEMP%")
    db_path = os.path.join(temp_directory, "database.db")
    cursor, connection = db_init(db_path)
    return cursor, connection

def sqlite_test():
    cursor, connection = sqlite_init()
    cursor.execute("DROP TABLE IF EXISTS pairs;")
    cursor.execute("CREATE TABLE IF NOT EXISTS pairs (key TEXT UNIQUE, value TEXT);")
    data = [ ("India", "New Delhi"), ("Japan" , "Tokyo"), ("USA", "Washington DC"), ("South Korea", "Seoul") ]
    cursor.executemany("INSERT INTO pairs (key, value) VALUES (?, ?);", data)
    # table = [ (key, value) for key, value in cursor.execute("SELECT key, value FROM pairs;").fetchall() ]
    table = [ key for key, in cursor.execute("SELECT key FROM pairs WHERE value != 'Tokyo';").fetchall() ]
    print(table)

def sqlite_test_2():
    cursor, connection = sqlite_init()
    cursor.execute("DROP TABLE IF EXISTS pairs;")
    cursor.execute("CREATE TABLE IF NOT EXISTS pairs (key TEXT UNIQUE ON CONFLICT REPLACE, value TEXT);")
    data = [ 
        ("condition_x", "True"), 
        ("condition_y", "True"), 
        ("condition_x", "False"), 
        ("condition_y", "False") 
    ]
    cursor.executemany("INSERT INTO pairs (key, value) VALUES (?, ?);", data)
    table = [ dict(row) for row in cursor.execute("SELECT * FROM pairs;").fetchall() ]
    ic(table)

def sqlite_test_3():
    cursor, connection = sqlite_init()
    cursor.execute("PRAGMA database_list")
    rows = [ dict(row) for row in cursor.fetchall() ]

def git_is_dirty_2():
    import git
    path = r"C:\Users\vivek\Desktop\bkp\scraper"
    repo = git.Repo(path)
    is_dirty = repo.is_dirty()
    breakpoint()
    ic(is_dirty)
    time.sleep(1)

def insert_dictionaries_test():
    from .db import insert_dictionaries
    cursor, connection = sqlite_init()

    path = r"C:\Users\vivek\Desktop\bkp\game"
    status = "CLEAN"
    url = "https://github.com/waivek/bkp_game"
    order = 0

    cursor.execute("DROP TABLE IF EXISTS paths;")
    cursor.execute("CREATE TABLE paths (path TEXT UNIQUE, url TEXT, 'order' INTEGER, status TEXT);")
    insert_dictionaries(cursor, "paths", [ { "path": path, "url": url, "status": status } ])
    insert_dictionaries(cursor, "paths", [ { "path": path, "order": order } ])
    insert_dictionaries(cursor, "paths", [ { "path": path, "order": order + 1} ])
    connection.commit()
    rows = [ dict(row) for row in cursor.execute("SELECT * FROM paths;").fetchall() ]
    ic(rows)

def print_string(string):
    for i in range(1000):
        print(f"[{i:02d}] {string} ")
        if string == "blue" and i == 3:
            message = f"Exception: string: {string}, i: {i}"
            raise Exception(message)
        time.sleep(1)


def start_many_threads():
    import threading
    colors = [ "red", "blue", "green", "yellow" ]
    thread_names = [ f"thread-{color}" for color in colors ]
    threads = [ threading.Thread(target=print_string, args=(color,), name=thread_name, daemon=False) for color, thread_name in zip(colors, thread_names) ]
    total = len(threads)
    for thread in threads:
        thread.start()
    while True:
        time.sleep(1)
        active_threads = [ thread for thread in threading.enumerate() if thread.name in thread_names ]
        count = len(active_threads)
        print(f"[start_many_threads] {count} / {total} alive")

def multiprocessing_multithreading_kill_test():
    from multiprocessing import Process
    from datetime import datetime
    from .error import print_error_information

    p1 = Process(target=start_many_threads)
    try:
        p1.start()
        TIMEOUT = 10
        for i in range(TIMEOUT, 0, -1):
            now = datetime.now()
            print(f"[{now:%H:%M:%S}] [multiprocessing] Going to terminate in {i} seconds ")
            time.sleep(1)
    except BaseException as e:
        print_error_information(e)
    finally:
        p1.terminate()

def internet(host="8.8.8.8", port=53, timeout=3):
    # 0.03s - 0.06s
    import socket
    """
    Host: 8.8.8.8 (google-public-dns-a.google.com)
    OpenPort: 53/tcp
    Service: domain (DNS/TCP)
    """
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(timeout)
        sock.connect((host, port))
        return True

    except socket.error as ex:
        return False

def f_one():
    cursor, connection = sqlite_init()
    cursor.execute("DROP TABLE IF EXISTS pairs;")
    cursor.execute("CREATE TABLE IF NOT EXISTS pairs (key TEXT UNIQUE, value TEXT);")
    data = [ ("India", "New Delhi"), ("Japan" , "Tokyo"), ("USA", "Washington DC"), ("South Korea", "Seoul") ]
    cursor.executemany("INSERT INTO pairs (key, value) VALUES (?, ?);", data)
    # rows = [ dict(row) for row in cursor.execute("SELECT * FROM pairs;").fetchall() ]
    # ic(rows)
    row = cursor.execute("SELECT count(*) max_order FROM pairs;").fetchone()
    breakpoint()

def sqlite_select_test():
    import os
    path = r"C:\Users\vivek\Documents\Python\backup-flask\data_dev.db"
    os.system(f'cmd /k sqlite3 {path} -cmd "SELECT * FROM paths"')

def print_loop():
    from datetime import datetime
    index = 0
    while True:
        index = index + 1
        ds = Code.LIGHTBLACK_EX + f"{datetime.now():%Y-%m-%d %H:%M:%S} :"
        print(f"{ds} {index}")
        time.sleep(1)

def path_test():
    import os.path
    path = r"C:\Users\vivek\Documents\Python\backup-flask\data_dev.db"
    L = os.path.splitext(path)
    ic(L)

def rich_movie():
    """Same as the table_movie.py but uses Live to update"""
    import time
    from contextlib import contextmanager

    from rich import box
    from rich.align import Align
    from rich.console import Console
    from rich.live import Live
    from rich.table import Table
    from rich.text import Text

    TABLE_DATA = [
        [
            "May 25, 1977",
            "Star Wars Ep. [b]IV[/]: [i]A New Hope",
            "$11,000,000",
            "$1,554,475",
            "$775,398,007",
        ],
        [
            "May 21, 1980",
            "Star Wars Ep. [b]V[/]: [i]The Empire Strikes Back",
            "$23,000,000",
            "$4,910,483",
            "$547,969,004",
        ],
        [
            "May 25, 1983",
            "Star Wars Ep. [b]VI[/b]: [i]Return of the Jedi",
            "$32,500,000",
            "$23,019,618",
            "$475,106,177",
        ],
        [
            "May 19, 1999",
            "Star Wars Ep. [b]I[/b]: [i]The phantom Menace",
            "$115,000,000",
            "$64,810,870",
            "$1,027,044,677",
        ],
        [
            "May 16, 2002",
            "Star Wars Ep. [b]II[/b]: [i]Attack of the Clones",
            "$115,000,000",
            "$80,027,814",
            "$656,695,615",
        ],
        [
            "May 19, 2005",
            "Star Wars Ep. [b]III[/b]: [i]Revenge of the Sith",
            "$115,500,000",
            "$380,270,577",
            "$848,998,877",
        ],
    ]

    console = Console()

    BEAT_TIME = 0.04


    @contextmanager
    def beat(length: int = 1) -> None:
        yield
        time.sleep(length * BEAT_TIME)


    table = Table(show_footer=False)
    table_centered = Align.center(table)

    console.clear()

    with Live(table_centered, console=console, screen=False, refresh_per_second=20):
        with beat(10):
            table.add_column("Release Date", no_wrap=True)

        with beat(10):
            table.add_column("Title", Text.from_markup("[b]Total", justify="right"))

        with beat(10):
            table.add_column("Budget", "[u]$412,000,000", no_wrap=True)

        with beat(10):
            table.add_column("Opening Weekend", "[u]$577,703,455", no_wrap=True)

        with beat(10):
            table.add_column("Box Office", "[u]$4,331,212,357", no_wrap=True)

        with beat(10):
            table.title = "Star Wars Box Office"

        with beat(10):
            table.title = (
                "[not italic]:popcorn:[/] Star Wars Box Office [not italic]:popcorn:[/]"
            )

        with beat(10):
            table.caption = "Made with Rich"

        with beat(10):
            table.caption = "Made with [b]Rich[/b]"

        with beat(10):
            table.caption = "Made with [b magenta not dim]Rich[/]"

        for row in TABLE_DATA:
            with beat(10):
                table.add_row(*row)

        with beat(10):
            table.show_footer = True

        table_width = console.measure(table).maximum

        with beat(10):
            table.columns[2].justify = "right"

        with beat(10):
            table.columns[3].justify = "right"

        with beat(10):
            table.columns[4].justify = "right"

        with beat(10):
            table.columns[2].header_style = "bold red"

        with beat(10):
            table.columns[3].header_style = "bold green"

        with beat(10):
            table.columns[4].header_style = "bold blue"

        with beat(10):
            table.columns[2].style = "red"

        with beat(10):
            table.columns[3].style = "green"

        with beat(10):
            table.columns[4].style = "blue"

        with beat(10):
            table.columns[0].style = "cyan"
            table.columns[0].header_style = "bold cyan"

        with beat(10):
            table.columns[1].style = "magenta"
            table.columns[1].header_style = "bold magenta"

        with beat(10):
            table.columns[2].footer_style = "bright_red"

        with beat(10):
            table.columns[3].footer_style = "bright_green"

        with beat(10):
            table.columns[4].footer_style = "bright_blue"

        with beat(10):
            table.row_styles = ["none", "dim"]

        with beat(10):
            table.border_style = "bright_yellow"

        for box_style in [
            box.SQUARE,
            box.MINIMAL,
            box.SIMPLE,
            box.SIMPLE_HEAD,
        ]:
            with beat(10):
                table.box = box_style

        with beat(10):
            table.pad_edge = False

        original_width = console.measure(table).maximum

        for width in range(original_width, console.width, 2):
            with beat(1):
                table.width = width

        for width in range(console.width, original_width, -2):
            with beat(1):
                table.width = width

        for width in range(original_width, 90, -2):
            with beat(1):
                table.width = width

        for width in range(90, original_width + 1, 2):
            with beat(1):
                table.width = width

        with beat(2):
            table.width = None

def ic_test():
    for i in range(1, 5):
        hundred = i * 20
        L = list(range(0, hundred))
        # print("---")
        # print()
        # print(L)
        # ic(L)
        # print()

def make_short(filename):
    D = { 
            r"C:\Users\vivek\AppData\Roaming\Python\Python310\site-packages": "site-packages",
            r"C:\Users\vivek\Documents\Python": "common",
            r"C:\Program Files\Python310\lib": "builtin",
    }
    for string, replacement in D.items():
        filename = filename.replace(string, replacement)
    return filename

f = None
def tracefunc(frame, event, arg, indent=[0]):
    global f
    f = frame
    if event == "call":
        indent[0] += 2
        filename = frame.f_code.co_filename
        short_filename = make_short(filename)
        if indent == [0]:
            print()
        print(" " * indent[0] + "  call function", frame.f_code.co_name, short_filename)
    elif event == "return":
        print(" " + " " * indent[0], "exit function", frame.f_code.co_name)
        indent[0] -= 2
    return tracefunc

def trace_on():
    import sys
    sys.setprofile(tracefunc)

def f_string():
    number = 3.2
    value = f"{number}"
    print(value)

def m_list(obj):
    match obj:
        case [x]:
            string = f"One Item: {x}"
        case [_, 4, _]:
            string = "Contains 4"
        case [x, y]:
            string = f"Two Items: ({x}, {y})"
        case [*items]:
            string = f"Many Items: {items}"
    print(string)
        

    # if type(obj) == dict and 'data' in obj:
    #     # helix-global, helix-streamer
    #     dictionaries = obj['data']
    #     table = [ { 'name': D['name'], 'id': D['id'], 'api': 'helix' } for D in dictionaries ]
    #     return table
    # # case { 'channel_emotes': y, **kwargs }:
    # if type(obj) == dict and 'channelEmotes' in obj:
    #     # bttv-streamer
    #     dictionaries = obj['channelEmotes'] + obj['sharedEmotes']
    #     table = [ { 'name': D['code'], 'id': D['id'], 'api': 'bttv' } for D in dictionaries ]
    #     return table

def parse_emote_response(obj):
    if obj == []:
        # https://api.betterttv.net/3/cached/frankerfacez/users/twitch/59299632
        return []
    # case { 'data': y, **kwargs }:
    match obj:
        case { 'data': dictionaries }:
            # helix-global, helix-streamer
            table = [ { 'name': D['name'], 'id': D['id'], 'api': 'helix' } for D in dictionaries ]
            return table
        case { 'channelEmotes': [*L1], 'sharedEmotes': [*L2] }:
            # bttv-streamer
            dictionaries = L1 + L2
            table = [ { 'name': D['code'], 'id': D['id'], 'api': 'bttv' } for D in dictionaries ]
            return table

    dictionaries = obj
    match dictionaries[0]:
        case { 'id': _, 'code': _, 'user': _ }:
            # ffz-global, ffz-streamer
            table = [ { 'name': D['code'], 'id': D['id'], 'api': 'ffz' } for D in dictionaries ]
            return table
        case { 'id': _, 'code': _ }:
            # bttv-global
            table = [ { 'name': D['code'], 'id': D['id'], 'api': 'bttv' } for D in dictionaries ]
            return table
        case { 'id': _, 'name': _ }:
            # stv-global, stv-streamer
            table = [ { 'name': D['name'], 'id': D['id'], 'api': 'stv' } for D in dictionaries ]
            return table


def m_dict(obj):
    match obj:
        case { 'data': dictionaries }:
            string = dictionaries
        case { 'channelEmotes': [*L1], 'sharedEmotes': [*L2] }:
            string = L1 + L2

    dictionaries = obj
    # if D.get('id') and D.get('code') and D.get('user') is not None:
    match dictionaries[0]:
        case { 'id':_, 'code':_, 'user': _ }:
            # ffz-global, ffz-streamer
            table = [ { 'name': D['code'], 'id': D['id'], 'api': 'ffz' } for D in dictionaries ]
            return table

        # if D.get('id') and D.get('code') and D.get('user') is None:
        case { 'id':_, 'code':_ }:
            # bttv-global
            table = [ { 'name': D['code'], 'id': D['id'], 'api': 'bttv' } for D in dictionaries ]
            return table


        # if D.get('id') and D.get('name'):
        case { 'id': _, 'name': _ }:
            # stv-global, stv-streamer
            table = [ { 'name': D['name'], 'id': D['id'], 'api': 'stv' } for D in dictionaries ]
            return table
    print(string)
        
def test_bp():
    print("Breakpoint!")
    breakpoint()


def test_handler():
    [ 1/0 for i in range(5) ]

def test_partition():
    from .common import create_partitions
    numbers = [ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ]
    pairs = create_partitions(len(numbers), 6)
    for start, end in pairs:
        list_slice = numbers[start:end]
        print(f"    {list_slice}")
    # end
    pass

def test_pathlib_ib():
    from pathlib import Path
    path = Path("C:/Users/file.txt")
    # ib(path)
    keys = [ key for key in dir(path) if not key.startswith("__") ]
    missing_attributes = []
    for key in keys:
        try:
            getattr(path, key)
        except Exception as e:
            missing_attributes.append(key)
    ic(missing_attributes)
    breakpoint()

def baz(string):
    return string.upper()

def bar(L):
    return ", ".join(str(item) for item in L)

def foo(x, y):
    return x+y

def rec(n):
    if n == 2:
        return
    return rec(n+1)

def test_trace():
    from .trace import trace
    from .trace import rerun
    trace([ "foo", "bar", "rec", "baz" ])
    ic(baz('latin'))
    ic(foo(1, 2))
    ic(bar(range(10)))
    ic(foo(3, 4))
    rec(0)
    rerun()

def get_frames():
    import sys
    import inspect
    inspect.stack()

    frames = []
    frame = sys._getframe()
    while frame:
        frames.append(frame)
        frame = frame.f_back
    ic(frames)

def test_sys():
    import sys
    L = [1, 2, 3]
    M = L + [ 4, 5, 6]
    print(sys.getrefcount(L))


# def dir():
#     from pathlib import Path
#     return Path(__file__).parent


def test_pathlib():
    timer.start("import os.path")
    import os.path; os.path
    t1 = timer.get("import os.path")
    timer.start("import Path")
    from pathlib import Path; Path
    t2 = timer.get("import Path")
    ic(t1, t2)
    from .reltools import here
    path = here() / "f1/f2/item.txt"
    path.parent.mkdir(parents=True, exist_ok=True)


def test_os_path():
    import os.path
    path = "/Users/../vivek/"

    ic(path)
    ic(os.path.normpath(path))
    ic(os.path.abspath(path))


def test_static_variable():
    
    if 'call_count' not in dir(test_static_variable):
        test_static_variable.call_count = 0

    test_static_variable.call_count = test_static_variable.call_count + 1

def test_color_bash():
    result = """
    SELECT * FROM items
    """
    print("".join([Code.LIGHTGREEN_EX + "waivek@vivobook", ":", Code.LIGHTBLUE_EX + "/mnt/c/Users/vivek/Documents/Python"]))

def test_yield_list(): 

    def db_gen(query):
        from .reltools import rel2abs
        from .db import db_init
        path = rel2abs("./db_test/gen.db")
        _, connection = db_init(path)

        while True:
            cursor = connection.cursor()
            yield cursor.execute(query)

    query = "SELECT * FROM items;"
    result = db_gen(query)
    ic(result)


def test_datenames(): 
    from datetime import datetime, timedelta
    now = datetime.now()
    day_table = []
    for daynumber in range(7):
        d1 = now - timedelta(days=daynumber)
        d2 = now - timedelta(days=daynumber+1)
        end_epoch = int(d1.timestamp())
        start_epoch = int(d2.timestamp())
        if daynumber == 0:
            dayname = "today"
        elif daynumber == 1:
            dayname = "yesterday"
        else:
            dayname = f"{d1 :%a}".lower()
        day_table.append({ 'dayname': dayname, 'start_epoch': start_epoch, 'end_epoch': end_epoch })
    range_D = { 
         "day": now - timedelta(hours=24),
         "week": now - timedelta(days=7),
         "month": now - timedelta(days=30),
         "year": now - timedelta(days=365),
         "all": datetime(year=2005, month=1, day=1)
    }
    range_table = [ { "rangename": rangename, "start_epoch": int(dt.timestamp()), "end_epoch": int(now.timestamp()) } for rangename, dt in range_D.items() ]
    return { "range": range_table, "day": day_table }


def simulate(expression):
    import sys
    from .frame import Frame
    frame = Frame(sys._getframe(1))
    start_line = frame.line.strip().replace("simulate(", "")[:-1]
    end_line = expression
    print(f">>> {start_line}")
    print(end_line)
    print()

def test_summaries():
    from .common import Date
    from .ic import ic; ic
    from datetime import datetime
    dt = datetime.now()
    year = dt.year
    date_string = f"{year}-01-01"
    date = Date(date_string)
    epoch = date.epoch
    epoch_string = f"Date({repr(date.epoch)})"
    str_string = f"Date({repr(date_string)})"
    strings = [epoch_string, str_string]
    # print("\n".join([ "date = " + string for string in  strings]))
    # ic(date.epoch)
    # ic(date.string)
    # ic(date.timeago())
    # ic(str(date))
    # ic(date.dt)

    # print()
    # print(">>> date.epoch")
    # print(f"{date.epoch}")
    # print()

    print(f">>> Date({repr(epoch)}) == Date({repr(date_string)})")
    print("True")
    print()
    print(f">>> date = Date({epoch})")
    simulate(date)
    simulate(date.epoch)
    simulate(date.string)
    simulate(date.timeago())
    simulate(str(date))
    simulate(date.dt)



# Vim Command: NTF --- call s:PythonNewTestFunction()
def main():
    test_summaries()
    # test_yield_list()
    # test_color_bash()
    # test_static_variable()
    # test_os_path()
    # test_pathlib()
    # test_pudb()
    # test_sys()
    # ic({ "a": 100, "b": 1, "c": 100_000_000})
    # test_trace()
    # test_pathlib_ib()
    # m_list([1, 4])
    # m_dict({ 'a': 'b', 'data': 'match_data', 'c': 'd' })
    # m_dict({ 'a': 'b', 'channelEmotes': 'cemotes', 'sharedEmotes': ['semotes'], 'c': 'd' })

if __name__ == "__main__":
    with handler():
        main()

timer.print("test.py")



from waivek import Timer   # Single Use
timer = Timer()
from waivek import Code    # Multi-Use
from waivek import handler # Single Use
from waivek import ic, ib     # Multi-Use, import time: 70ms - 110ms
Code; ic; ib; handler

def main():
    pass

if __name__ == "__main__":
    with handler():
        main()



# IMPORTANT:
# 1. FILE HAS TO BE IN THE SAME DIRECTORY AS COMMON IF YOU ARE IMPORTING THROUGH COMMON
# 2. get_caller_parent HAS TO BE IN SAME FILE AS rel2abs

# TODO:
# test stack deep-ness in 
#
#     reltools:write
#     reltools:read
#     db:db_init`

import os
import sys
from .color import Code

def pathjoin(frame, relpath):
    frame_directory = os.path.dirname(frame.f_code.co_filename)
    return os.path.realpath(os.path.join(frame_directory, relpath))

def readlines(relpath):
    """
    Read lines from a file.
    relpath can be relative or absolute. Resolved via rel2abs methodology.
    Leading and trailing whitespace is stripped.
    """
    path = relpath if os.path.isabs(relpath) else pathjoin(sys._getframe(1), relpath)
    if os.path.exists(path) is False:
        print(f"File not found: {Code.RED + path}")
    with open(path, "rb") as f:
        lines = f.read().decode("utf-8").strip().splitlines()
    return lines
    
def writelines(relpath, lines):
    """
    Write lines to a file.
    relpath can be relative or absolute. Resolved via rel2abs methodology.
    """
    path = relpath if os.path.isabs(relpath) else pathjoin(sys._getframe(1), relpath)
    with open(path, "wb") as f:
        f.write("\n".join(lines).encode("utf-8"))

def read(relpath):
    import json
    from .color import Code
    filepath = relpath if os.path.isabs(relpath) else pathjoin(sys._getframe(1), relpath)
    filepath = os.path.normpath(filepath)
    if os.path.exists(filepath) is False:
        message = "[reltools.py:read()] " + (Code.YELLOW + "File does not exist: ") + (Code.RED + filepath)
        print(message)
        return {}

    with open(filepath, "r") as f:
        obj = json.load(f)
    return obj

def write(obj, relpath):
    import json
    filepath = os.path.normpath(rel2abs(relpath))
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    with open(filepath, "w") as f:
        json.dump(obj, f, indent=4)
    return filepath

def get_caller_parent():
    from .frame import Frame
    self_path = os.path.normpath(os.path.dirname(__file__))
    frame_index = 0
    directories = []
    frames = []
    while True:
        frame = sys._getframe(frame_index)
        frames.append(Frame(frame))
        called_path      = frame.f_code.co_filename
        parent_directory = os.path.dirname(os.path.realpath(called_path))
        directories.append(parent_directory)
        if parent_directory.lower() != self_path.lower():
            break
        if frame.f_back == None:
            break
        frame_index = frame_index + 1


    if frame.f_back is None and len(set(directories)) == 1 and directories[0] == os.path.dirname(__file__): # importing from a python file in same directory as reltools.py
        return directories[0]

    return parent_directory

def frame_to_absolute_directory(frame):
    return os.path.dirname(os.path.realpath(frame.f_code.co_filename))

def rel2abs(relative_path):
    from .ic import ic
    from .ic import ib
    ic; ib
    return pathjoin(sys._getframe(1), relative_path)

    # Implementation 3
    # ================
    parent_directory = os.path.dirname(sys._getframe(1).f_globals["__file__"])
    absolute_path = os.path.realpath(os.path.join(parent_directory, relative_path))
    ic(absolute_path)
    return absolute_path

    # self_function_name = sys._getframe(0).f_code.co_name

    # Implementation 2
    # ================
    frame = sys._getframe(1)
    parent_directory = frame_to_absolute_directory(frame)
    absolute_path    = os.path.realpath(os.path.join(parent_directory, relative_path))
    ic(absolute_path)


    # Implementation 1
    # ================
    # parent_directory = get_caller_parent()
    # absolute_path    = os.path.realpath(os.path.join(parent_directory, relative_path))
    # ic(absolute_path)

    return absolute_path

def here():
    # importing pathlib is slow as shit
    # return Path(sys._getframe(1).f_code.co_filename).parent
    return os.path.dirname(os.path.realpath(sys._getframe(1).f_code.co_filename))




# See `IMPORT INFORMATION` section in ic()


# Brief:
# 1. Timing things
# 2. Printing Error messages
# 3. Printing variables
# 4. Printing dictionaries and tables
# 5. Autocomplete
# 6. Coloring Strings <externalized to color.py>

from .timer import Timer
timer = Timer()

from .common import print_dict, truncate, Date
from .color import Code

import sys
import os
import os.path
import types

# TextDoc.docroutine

# docroutine {{{
def classname(object, modname):
    """Get a class name and qualify it with a module name if necessary."""
    name = object.__name__
    if object.__module__ != modname:
        name = object.__module__ + '.' + name
    return name

def getdoc(object):
    import re
    import inspect
    """Get the doc string or comments for an object."""
    result = inspect.getdoc(object) or inspect.getcomments(object)
    return result and re.sub('^ *\n', '', result.rstrip()) or ''

def indent(text, prefix='    '):
    """Indent text by prepending a given prefix to each line."""
    if not text:
        return ''
    lines = [prefix + line for line in text.split('\n')]
    if lines:
        lines[-1] = lines[-1].rstrip()
    return '\n'.join(lines)

def _is_bound_method(fn):
    import inspect
    if inspect.ismethod(fn):
        return True
    if inspect.isbuiltin(fn):
        self = getattr(fn, '__self__', None)
        return not (inspect.ismodule(self) or (self is None))
    return False

def docroutine(object, name=None, mod=None, cl=None):
    import inspect
    """Produce text documentation for a function or method object."""
    realname = object.__name__
    name = name or realname
    note = ''
    skipdocs = 0
    if _is_bound_method(object):
        imclass = object.__self__.__class__
        if cl:
            if imclass is not cl:
                note = ' from ' + classname(imclass, mod)
        else:
            if object.__self__ is not None:
                note = ' method of %s instance' % classname(
                    object.__self__.__class__, mod)
            else:
                note = ' unbound %s method' % classname(imclass, mod)

    if name == realname:
        title = realname
    else:
        if cl and realname in cl.__dict__ and cl.__dict__[realname] is object:
            skipdocs = 1
        title = name + ' = ' + realname
        
    argspec = None

    if inspect.isroutine(object):
        try:
            signature = inspect.signature(object)
        except (ValueError, TypeError):
            signature = None
        if signature:
            argspec = str(signature)
            if realname == '<lambda>':
                title = name + ' lambda '
                # XXX lambda's won't usually have func_annotations['return']
                # since the syntax doesn't support, but it is possible.
                # So removing parentheses isn't truly safe.
                argspec = argspec[1:-1]  # remove parentheses
    if not argspec:
        argspec = '(...)'
    decl = title + argspec + note

    if skipdocs:
        return decl + '\n'
    else:
        doc = getdoc(object) or ''
        return '\n' + decl + '\n' + (doc and indent(doc).rstrip() + '\n')
    # }}}


def sig(frame_index=1):
    import inspect

    frame_stack = inspect.stack()
    frame_info = frame_stack[frame_index]
    frame = frame_info[0]

    fname = frame_info.function

    # Works in Python 3.8 but not 3.10
    # import gc
    # function_object = gc.get_referrers(frame.f_code)[0]

    if fname != "<module>":
        function_object = frame.f_globals[fname]
        keys = inspect.getfullargspec(function_object).args
    else:
        keys = []

    # fname = frame_info.function
    # function_object = eval(fname)
    # keys = inspect.getfullargspec(function_object).args

    D = frame.f_locals
    values = [ D[key] for key in keys ]
    parameter_string = "(" + ", ".join(repr(value) for value in values) + ")"
    value_string = "{function_name}{parameters}".format(function_name=fname, parameters=parameter_string)

    return value_string

def get_context(callFrame=None):
    import inspect
    callFrame = callFrame if callFrame else sys._getframe().f_back
    frameInfo = inspect.getframeinfo(callFrame)
    # lineNumber = frameInfo.lineno
    lineNumber = callFrame.f_lineno
    # parentFunction = frameInfo.function
    filename = os.path.basename(frameInfo.filename)
    # if parentFunction != '<module>':
    #     parentFunction = '%s()' % parentFunction
    signature = sig(3)
    return '%s:%s %s' % (filename, lineNumber, signature)

def is_function(value):
    function_types = [
        types.BuiltinFunctionType,
        types.FunctionType,
        types.MethodDescriptorType,
        types.MethodType,
    ]
    result = type(value) in function_types
    return result

#  tuple_table:
#     1. [ tuple | list ]
#     2. all sublists are identical in length
# single_dict_table:
#     1. type == dict
# multi_dict_table:
#     1. [ dict ]
#     2. all dicts are identical in keys()

def is_tuple_table(table):
    if type(table) != list:
        return False
    types = list(set(type(row) for row in table))
    if len(types) != 1:
        return False
    common_type = types[0]
    return common_type in [ tuple, list ]

def is_single_dict_table(table):
    # we add items check so that Date class doesn’t pass
    return type(table) == dict or (hasattr(table, '__dict__') and hasattr(table, 'items'))

def is_multi_dict_table(table):
    if type(table) != list:
        return False
    if any(type(row) != dict for row in table):
        return False
    from itertools import groupby
    unique_key_combinations = list(groupby([ D.keys() for D in table ]))
    return len(unique_key_combinations) == 1

def table_friendly(table):
    return is_tuple_table(table) or is_single_dict_table(table) or is_multi_dict_table(table)

# is_tuple_table, is_single_dict_table, is_multi_dict_table
def list_fmt(L):
    max_width, max_height = os.get_terminal_size()
    max_height = int(os.get_terminal_size().lines / 2)
    half_height = int(max_height / 2)
    list_length = len(L)
    space_count = len(str(list_length)) + 2
    max_width = max_width - space_count - 5
    if len(str(L)) < max_width:
        return str(L)
    elif list_length > max_height:
        lines = []
        for i in range(0, half_height):
            L_string = truncate(str(L[i]), max_width)
            padded_i = f"[{i}]".rjust(space_count)
            lines.append(f"{padded_i}: {L_string}")
        lines.append(f"{' ' * space_count}  ...")
        for i in range(list_length - half_height, list_length):
            L_string = truncate(str(L[i]), max_width)
            padded_i = f"[{i}]".rjust(space_count)
            lines.append(f"{padded_i}: {L_string}")
        return "\n".join(lines)
    else:
        lines = []
        for i in range(0, list_length):
            L_string = truncate(str(L[i]), max_width)
            padded_i = f"[{i}]".rjust(space_count)
            lines.append(f"{padded_i}: {L_string}")
        return "\n".join(lines)

def describe_function(arg):
    name = getattr(arg, '__name__', None)
    doc_string = docroutine(arg, name)
    return doc_string

def ic_one(value):
    if table_friendly(value):
        table = Table()
        if is_tuple_table(value) or is_multi_dict_table(value):
            for row in value:
                table.row(row)
        else:
            table.parse(value)
        return str(table)
    if type(value) == list:
        return list_fmt(value)
    if type(value) == Date:
        # import dateutil.parser
        from datetime import timezone
        table = Table()
        table.parse({
            "ist": value.string,
            "utc": value.dt.astimezone(timezone.utc).isoformat()[:-6],
            "epoch": value.epoch,
            "relative": value.timeago()
        })
        return str(table)
    if is_function(value):
        doc_string = describe_function(value)
        return doc_string
    if type(value) == types.ModuleType:
        return ic_one({
            key: getattr(value, key) .__doc__.strip().split("\n")[0] if getattr(value, key).__doc__ else "No __doc__"
            for key in dir(value) if is_function(getattr(value, key)) and not key.startswith("_") 
        })
    return str(value)

def ib(obj):
    def custom_str(v):
        if is_function(v):
            return v.__doc__.strip().split("\n")[0] if v.__doc__ else "No __doc__"
        else:
            return truncate(str(v), 160)
    keys = [ key for key in dir(obj) if not key.startswith("__") ]

    keys.sort(key=lambda key: not is_function(getattr(obj, key)))
    D = { key : custom_str(getattr(obj, key)) for key in keys }
    print_dict(D)

def inspect_dot_stack():
    frame_index = 1
    frames = []
    while True:
        try:
            frame = sys._getframe(frame_index)
        except ValueError: 
            break
        frames.append(frame)
        frame_index = frame_index + 1
    return frames

def pdb_check():
    # import inspect
    # stack = inspect.stack() # Takes 300 ms
    # frames = [ frame_info.frame for frame_info in stack ]

    frames = inspect_dot_stack()
    frame_names = [ frame.f_globals['__name__'] for frame in frames ]
    if 'pdb' in frame_names or 'pydevd' in frame_names:
        return True
    else:
        return False
    # modules = [ inspect.getmodule(frame) for frame in frames ]
    # return pdb in modules

path_to_lines = {}
def frame_to_line(frame):
    # Implementation 1
    # ================
    # import linecache
    # linecache.getline(frame.f_code.co_filename, frame.f_lineno).strip()
    path = frame.f_code.co_filename
    if path not in path_to_lines:
        with open(path, "r") as f:
            lines = f.readlines()
        path_to_lines[path] = lines
    lines = path_to_lines[path]
    line = lines[frame.f_lineno-1]
    return line

def ic(*values):
    # IMPORT INFORMATION
    # ==================
    #
    # ast     : Only import ast for instances like ic(1, 2) or ic(foo(a), [1,2,3]) and not of ic(1) ic(2) which is most common use case
    # inspect : Only import for ic() / ic(function) / ic(module) / sig()

    in_pdb = pdb_check()

    if not values:
        if in_pdb:
            print(Code.GREEN + "IN (pdb).")
            return
        callFrame = sys._getframe().f_back
        context_string = get_context(callFrame)
        print(context_string)
        return

    x = values[0]
    if len(values) == 1 and ic_one(x) != str(x):
        # print("A")
        print(ic_one(x))
        return

    if in_pdb:
        for value in values:
            print(str(value))
        return


    # Implementation 3:
    # =================
    frame = sys._getframe(1)
    line = frame_to_line(frame).strip()
    if len(values) == 1:
        print(f"{line[3:-1]}: {Code.LIGHTCYAN_EX+values[0]}")
    else:
        import ast
        node = ast.parse(line).body[0].value
        for arg, value in zip(node.args, values):
            print(f"{ast.unparse(arg)}: {Code.LIGHTCYAN_EX+value}")


__all__ = [ "ic", "ib" ]


def len_without_ansi_codes(s):
    import re
    ansi_codes = [ '\x1b[30m', '\x1b[31m', '\x1b[32m', '\x1b[33m', '\x1b[34m', '\x1b[35m', '\x1b[36m', '\x1b[37m', '\x1b[90m', '\x1b[91m', '\x1b[92m', '\x1b[93m', '\x1b[94m', '\x1b[95m', '\x1b[96m', '\x1b[97m' ]
    ansi_codes_joined = "".join(ansi_codes)
    return len(re.sub(f'[{ansi_codes_joined}]', '', s))

def log(dictionaries):
    from .db import db_init
    from time import time
    import json
    cursor, connection = db_init("errors/ic.db")
    cursor.execute("CREATE TABLE IF NOT EXISTS errors (epoch INTEGER NOT NULL, list_json TEXT NOT NULL UNIQUE ON CONFLICT IGNORE);")
    list_json = json.dumps(dictionaries)
    epoch = int(time())
    cursor.execute("INSERT INTO errors (epoch, list_json) VALUES (?, ?);", (epoch, list_json))
    connection.commit()

def save_ic_table_error(dictionaries):
    from datetime import datetime
    from .reltools import write
    dt = datetime.now()
    filename = f"{dt:%y%m%d-%Hh%Mm%Ss}.json"
    path = os.path.abspath(os.path.expanduser(f'~/Documents/Python/ic-test-manual/{filename}'))
    write(dictionaries, filename)

class Table:
    # Why we can’t use columnar
    # 
    # 1. No option for separator only between columns. Any separator you choose is also part of the border
    # 2. 150ms import time vanilla 
    def __init__(self):
        self.gutter = '  '
        self.separator = '  '
        self.table = []

    def row(self, *args):
        if len(args) == 1:
            L = args[0]
        else:
            L = list(args)
        self.table.append(L)

    def parse(self, D):
        self.table = D

    def normalize_table(self):
        table = self.table
        if is_tuple_table(table):
            L = [ tuple(sublist) for sublist in table ]
            return [], L
        elif is_single_dict_table(table):
            items = list(table.items())
            return [ "KEY", "VALUE" ], items
        elif is_multi_dict_table(table):
            keys = table[0].keys()
            L = [ tuple(D[key] for key in keys) for D in table ]
            return keys, L
        return None

    def get_column_tuples(self, headers, tuple_table):
        def get_column_just(column):
            is_numeric = all(type(cell) in [int, float] for cell in column)
            return str.rjust if is_numeric else str.ljust
        def get_width(cell):
            # Handles multi-line strings
            cell_str = str(cell)
            # return max(len_without_ansi_codes(line) for line in cell_str.split("\n"))
            return max(len(line) for line in cell_str.split("\n"))

        column_count = len(tuple_table[0])
        columns = [ list(row[i] for row in tuple_table) for i in range(column_count) ]
        column_tuples = [ ( max(get_width(cell) for cell in column), get_column_just(column) ) for column in columns  ]
        if headers:
            column_tuples = [ (max(width, len(header)), just) for header, (width, just) in zip(headers, column_tuples) ]

        # Distribute Column Widths given terminal size
        initial_widths = [ width for width, _ in column_tuples ]
        terminal_width, _ = os.get_terminal_size()
        gutter_width = len(self.gutter) * 2
        separator_width = len(self.separator) * (column_count-1)
        column_width_total = terminal_width - gutter_width - separator_width
        min_column_width = int(column_width_total / column_count)
        small_indices = [ i for i, width in enumerate(initial_widths) if width <= min_column_width ]
        big_indices = [ i for i, width in enumerate(initial_widths) if width > min_column_width ]
        if len(big_indices) == 0:
            return column_tuples
        small_width_total = sum(width for width in initial_widths if width <= min_column_width)
        remaining_width_total = column_width_total - small_width_total
        big_width = int(remaining_width_total / len(big_indices))
        remaining_width_total = remaining_width_total - big_width * len(big_indices)
        final_big_width = big_width + remaining_width_total
        for i in big_indices:
            _, just = column_tuples[i]
            column_tuples[i] = (big_width, just)
            if i == len(big_indices)-1:
                column_tuples[i] = (final_big_width, just)

        final_widths = [ width for width, _ in column_tuples ]
        calc_total_width = gutter_width + separator_width + sum(final_widths)
        if calc_total_width != terminal_width:
            log(self.table)
            print(Code.RED +  "calc_total_width != terminal_width")
            print(Code.RED + f"           ({calc_total_width}) !=          ({terminal_width})")
            print(f"Initial Widths                  : {initial_widths}")
            print(f"Column Width All                : {column_width_total}")
            print(f"Minimum Column Width            : {min_column_width}")
            print(f"Final  Widths                   : {final_widths}")
            print("Do 'u' and then run `save_ic_table_error(x)`")
            breakpoint()
        return column_tuples

    def wrap_box(item, width, height, just):
        # Purpose: Handle truncate, wrap, newline, padding
        # If there is a newline, avoid truncate() probably and truncate on line
        # Make sure returned result is padded correctly and all lines have the width specified
        # If >1 line, all lines HAVE to have the same width
        #
        # For really big dicts and lists do two things 
        # 1.1) color code keys/top-level commas in the box so that it is easy to parse
        # 1.2) give alternating colors to top-level items so that it is easy to parse
        # 2) use as much horizontal space possible
        # 
        # Handle color codes in wrapped lienes
        # 1) Right now color codes are counted in the length of the file (this is in `Table.get_width` and `textwrap.wrap`)
        # 2) I'm pretty sure color codes spill over to next cell. Illusion only works as long as colored-block is also the last block
        #
        # Ideally ic should not insert any new color on it’s on. Input’s should come in colored and ic resets multi-line color blocks appropriately
        # ic should also not replace a "\n" with a r"\n". This should also be done before ic()
        #
        # There is something very appealing about ic printing by default in black-and-white AS IS
        #
        # truncate() also has to take into account length of ANSI escape characters
        #

        pass

    def to_string(self):
        def fmt(cells, column_tuples):
            S = self.separator
            G = self.gutter
            all_strings_are_single_line = all("\n" not in str(cell) for cell in cells)
            if all_strings_are_single_line:
                return G + S.join(just(truncate(str(cell), width), width) for cell, (width, just) in zip(cells, column_tuples))
            else:
                cells = [ str(cell) for cell in cells ]
                largest_height = max(len(cell.split("\n")) for cell in cells)
                v_pad_cells = [ v_pad(cell.split("\n"), largest_height) for cell in cells ]
                return_lines = []
                for i in range(largest_height):
                    parts = [ v_pad_cell[i] for v_pad_cell in v_pad_cells ]
                    parts = [ just(truncate(part, width), width) for part, (width, just) in zip(parts, column_tuples) ] 
                    return_lines.append(G + S.join(parts))
                return "\n".join(return_lines)

        headers, tuple_table = self.normalize_table()
        column_tuples = self.get_column_tuples(headers, tuple_table)

        header_lines = [ "", fmt(headers, column_tuples), "" ] if headers else [ "" ]
        footer_lines = [ "" ]
        row_lines = [ fmt(row, column_tuples) for row in tuple_table ]
        table_string = "\n".join(header_lines+row_lines+footer_lines)
        return table_string

    def __str__(self):
        return self.to_string()

# test_ic (8 functions) {{{
def data_source_paths():
    T = [ 
        (r"C:\users\vivek\Documents\Python\ic.py:163", r"result = upper / lower"),
        (r"C:\users\vivek\Documents\Python\ic.py:229", r"ic_one(value)"),
        (r"C:\Users\vivek\Desktop\Twitch\coffee-vps\code\twitch_utilities.py:4112", r"ic(st)"),
        (r"C:\Users\vivek\Desktop\Twitch\coffee-vps\code\twitch_utilities.py:4225", r"main()") 
    ]
    return T

def data_source_mixed_dict():
    from datetime import datetime
    D = {
        "epoch": 1641749543,
        100: "hundred",
        "date": datetime.now(),
        "list": [1, 2, 3, 4]
    }
    return D

# Required for error.print_variables_by_frame
def data_source_multiline_long_colored_variables():
    # import pdb
    # tall_string = str(pdb.__doc__)
    import textwrap
    from .error import color_D_if_big
    range_count = 100
    numbers = list(range(range_count))
    alphabets = [ "abcdefghijklmnopqrstuvwxyz", "ABCDEFGHIJKLMNOPQRSTUVWXYZ" ]
    select_sql_extra_space = r"""
    SELECT datetime(created_at, "+"||offset||" seconds") clip_utc FROM 
        (SELECT id, created_at FROM videos)
        INNER JOIN
        (SELECT slug, "vod.id", "vod.offset" offset FROM kraken_clips)
    ON id="vod.id";
    """
    select_sql = textwrap.dedent(select_sql_extra_space).strip()

    # local_D = locals()
    # mystuff = "\n".join(textwrap.wrap(color_D_if_big(local_D), 80))
    from rich import ansi
    decoder = ansi.AnsiDecoder()


    timer.start("import Table")
    from rich.console import Console
    from rich.table import Table
    timer.print("import Table")
    table = Table(title="two-column display")
    mystuff = next(decoder.decode(color_D_if_big(locals())))
    table.add_column("key")
    table.add_column("value")
    table.add_column("type")
    for key, value in locals().items():
        if key == "mystuff":
            table.add_row(key, value, str(type(value)))
        else:
            table.add_row(key, str(value), str(type(value)))
    console = Console()
    console.print(table)
    # r_print(local_D["mystuff"])




    # table = Table()
    # for key, value in locals().items():
    #     if key == 'table':
    #         continue
    #     table.row(key, str(value), str(type(value)))
    # print(table)


    

def single_dict_table():
    D = data_source_mixed_dict()
    return D

def multi_dict_table():
    T = data_source_paths()
    return [ { "path": path, "line": line } for path, line in T ]

def tuple_table():
    return data_source_paths()

def fn_call():
    return 'Called Result'

def joined_string(prefix, postfix, D, call_result, da):
    return " | ".join([ prefix, postfix, str(D), call_result, str(da)])

def test_ic():
    table1 = tuple_table()
    table2 = multi_dict_table()
    table3 = single_dict_table()

    ic(table1)
    ic(table2)
    ic(table3)
    ic(os.path)
    ic(Date.now())

    print()

    ret = "Hello, World"
    before = 'prefix '
    ic()
    ic(joined_string(before, '!', {'cat': 'dog'}, fn_call(), Date.now()))
    ic(ret == before)
    ic_locals = locals()
    ic_locals = { str(key).strip(): str(value).strip() for key, value in ic_locals.items() }
    ic(ic_locals)
    print()


    # }}}

# IntelliJ: Replace VCS Buttons
# IntelliJ: Button at Top
# IntelliJ: Go To Line
# IntelliJ: Breakpoint Functionality
# IntelliJ: Structure Functionality

# [DONE] Vim: https://github.com/jeetsukumaran/vim-pythonsense#stock-vim-vs-pythonsense-motions
# [DONE] Vim: https://github.com/wellle/context.vim
# [DONE] Vim: Jedi Intellisense
# [DONE] Vim: gd for class
# [DONE] Vim: go to end of function 


def table_print():
    #  data
    #  headers
    #  head = 0
    #  justify = 'l'
    #  wrap_max = 5
    #  max_column_width = None
    #  min_column_width = 5
    #  row_sep = '-'
    #  column_sep = '|'
    #  patterns = []
    #  drop = []
    #  select = []
    #  no_borders = False
    #  terminal_width = None
    from columnar import columnar
    no_top_padding = True
    table = [ list(T) for T in tuple_table() ]
    table[0][1] = Code.GREEN + table[0][1]
    result = columnar(table, row_sep="", column_sep="...")
    if no_top_padding:
        result = "\n".join(result.split("\n")[1:])

    print("\n-- START --")
    print(result)
    print("-- FINISH --")

def v_pad(list_or_string, line_count=1):

    L = [ list_or_string ] if type(list_or_string) == str else list_or_string
    max_length = max(len(s) for s in L)
    list_height = len(L)
    if list_height <= line_count:
        blank_line_count = line_count - list_height
        result_lines = L + [ " " * max_length ] * blank_line_count
    else:
        top_end = int(line_count/2)
        bot_start = list_height - int(line_count/2) + 1
        slice_L = L[0:top_end] + [ "..." ] + L[bot_start:]
        if len(slice_L) != line_count:
            print(Code.RED +  "(error) v_pad(list_string, line_count)")
            print(Code.RED + f"....... len(slice_L): {len(slice_L)}")
            print(Code.RED + f"....... line_count  : {line_count}")
            breakpoint()
        result_lines = slice_L
    return result_lines

def wrap_test():
    # multi-line variable in table support
    # use case: C:\Users\vivek\Desktop\Twitch\coffee-vps\code\err_test.py
    import textwrap
    import pdb
    target_width = 70
    # target_height = 10
    range_count = 100

    list_string = str(list(range(range_count)))
    # list_string = "".join(str(i) for i in range(range_count))

    small_string = f"Numbers till {range_count}"

    tall_string = str(pdb.__doc__).strip()

    list_parts = textwrap.wrap(list_string, target_width)
    list_parts = [ list_part.ljust(target_width) for list_part in list_parts ]
    target_height = len(list_parts)
    small_parts = v_pad(small_string, target_height)
    tall_parts = v_pad(tall_string.split("\n"), target_height)

    single_toggle = True
    for small_part, list_part, tall_part in zip(small_parts, list_parts, tall_parts):
        if single_toggle:
            print(small_part + " ... " + list_part + " ... " + tall_part)
            single_toggle = False
        else:
            print(small_part + "     " + list_part + "     " + tall_part)

def foo(x, y):
    return x + y


def baz(string):
    return string.upper()

def get_args(*values):
    import executing
    # import inspect
    import ast
    stack = inspect_dot_stack()
    frame = sys._getframe(1)
    
    node = executing.Source.executing(frame).node
    print(ast.unparse(node))
    import inspect
    import linecache
    source = inspect.getsource(frame)
    mod = ast.parse(''.join(linecache.getlines(frame.f_code.co_filename)))
    breakpoint()
    # nodes = ast.parse(source)
    # for node in ast.walk(nodes):
    #     try:
    #         if node.value.func.id == 'get_args':
    #             print("---")
    #             print(ast.unparse(node))
    #     except:
    #         pass

    print(frame.f_lineno)
    # print(ast.unparse(nodes))
    print()


def error_1():
    from .db import db_init
    import json
    cursor, connection = db_init("errors/ic.db")
    dictionaries = [ dict(row) for row in cursor.execute("SELECT * FROM errors") ]
    json_strings = [ D["list_json"] for D in dictionaries ]
    dictionaries = [ json.loads(json_string) for json_string in json_strings ]
    for D in dictionaries:
        ic(D)

def error_2():
    from reltools import read
    D = read("ic-test-manual/221015.json")
    ic(D)


def main():
    error_2() # Doesn’t Work / Reproduce the error
    return 
    save_ic_table_error([])
    return
    # ic(1)
    string = '1'
    # ic(string)
    # ic(foo(1,2), 3)
    # get_args(foo(1, 2), 'Three')
    return
    string = 'Hey'
    get_args(baz(string))
    return
    # _visit_after_children: C:\Users\vivek\AppData\Roaming\Python\Python37\site-packages\asttokens\mark_tokens.py :63
    # test_ic()
    # calc_len = len_without_ansi_codes(s)
    # normal_len = len(s)
    # ic(calc_len, normal_len)
    # # Required for error.print_variables_by_frame
    # # data_source_multiline_long_colored_variables()

# main()

timer.no_print = False

if __name__ == "__main__":

    from .error import handler
    with handler():
        main()




def get_pairs(pairs_type):
    if pairs_type == bytes:
        pairs = [ ("nothing", None), ("binary", "Hello, World!".encode("utf-8") ), ("string", "b") ]
    elif pairs_type == list:
        numbers = list(range(10))
        pairs = [ ("nothing", None), ("numbers", numbers ), ("empty", None) ]
    return pairs

def test_ic_custom_types():

    from .db import db_init, insert_dictionaries, get_schema
    from .ic import ic
    cursor, connection = db_init("db_test/lists.db")
    cursor.execute("DROP TABLE IF EXISTS pairs;")

    pairs = get_pairs(list)

    dictionaries = [ { "key": key, "value": value } for key, value in pairs ]
    # cursor.execute("CREATE TABLE pairs ([key] TEXT, value LIST);")
    insert_dictionaries(cursor, "pairs", dictionaries)
    # cursor.executemany("INSERT INTO pairs ([key], value) VALUES (:key, :value);", dictionaries)
    connection.commit()
    dictionaries = [ { "key": key, "key_type": type(key), "value": value, "key_value": type(value) } for key, value in cursor.execute("SELECT key, value FROM pairs;") ]
    ic(dictionaries)
    print(f"CREATE TABLE pairs ({get_schema(cursor, 'pairs').column_definitions()});")





from .color import Code
from .common import create_partitions, smart_pad, enumerate_count
from .common import Timestamp, Date
from .db import db_init, insert_dictionaries
from .error import handler
from .get import aget
from .ic import ic, ib
from .print_utils import head, truncate, abbreviate
from .reltools import rel2abs, read, write, readlines, writelines
from .timer import Timer





# Hello
# Red
# Blue
from .timer import Timer
timer = Timer()
from .common import print_red_line, make_string_green, truncate, Timer
from .reltools import pathjoin
from enum import IntEnum
timer = Timer()

import sys

if sys.platform != "win32":
    import pysqlite3 as sqlite3
else:
    import sqlite3

def table_exists(cursor, table_name):
    statement = f"SELECT * FROM sqlite_master WHERE type='table' AND name='{table_name}'"
    D = cursor.execute(statement).fetchall()
    return True if D else False

def dump_list(L):
    # print("dump_list", L)
    import json
    if L is None:
        return '[]'
    else:
        return json.dumps(L)

def load_json_bytes(json_bytes):
    print("load_json_bytes", json_bytes, type(json_bytes))
    import json
    if json_bytes is None:
        return []
    else:
        return json.loads(json_bytes)


def db_init(db_path):
    sqlite3.register_adapter(list, dump_list)
    sqlite3.register_converter("LIST", load_json_bytes)

    import os
    db_path = pathjoin(sys._getframe(1), db_path)
    os.makedirs(os.path.dirname(db_path), exist_ok=True)
    connection = sqlite3.connect(db_path, detect_types=sqlite3.PARSE_DECLTYPES)
    # connection = sqlite3.connect(db_path, detect_types=sqlite3.PARSE_COLNAMES)
    connection.row_factory = sqlite3.Row
    cursor = connection.cursor()
    return cursor, connection

def jsonify_lists(D):
    import json
    list_keys = [ key for key, value in D.items() if type(value) == list ]
    for key in list_keys:
        L = D[key]
        D[key] = json.dumps(L)
    return D

# flatten_dict({ "vim": "god", "css": "okay" }, "skills")
#     ->
# [ ("skills.vim", "god"), ("skills.css", "okay") ]
def flatten_dict(D, parent_keys=[]):
    pairs = []
    for k, v in D.items():
        if type(v) == dict:
            pairs = pairs + flatten_dict(v, parent_keys+[k])
        else:
            new_k = ".".join(parent_keys+[k])
            pairs.append((new_k, v))
    return pairs


def prepare_dict(D):
    pairs = flatten_dict(D)        # 1.10, 60%
    D = { k: v for k, v in pairs } # 0.26, 15%
    # D = jsonify_lists(D)           # 0.45, 25%
    return D

def is_reserved(key):
    # 147 Words "ABORT" ... "WITHOUT"
    sqlite_reserved_words = [ "ABORT", "ACTION", "ADD", "AFTER", "ALL", "ALTER", "ALWAYS", "ANALYZE", "AND", "AS", "ASC", "ATTACH", "AUTOINCREMENT", "BEFORE", "BEGIN", "BETWEEN", "BY", "CASCADE", "CASE", "CAST", "CHECK", "COLLATE", "COLUMN", "COMMIT", "CONFLICT", "CONSTRAINT", "CREATE", "CROSS", "CURRENT", "CURRENT_DATE", "CURRENT_TIME", "CURRENT_TIMESTAMP", "DATABASE", "DEFAULT", "DEFERRABLE", "DEFERRED", "DELETE", "DESC", "DETACH", "DISTINCT", "DO", "DROP", "EACH", "ELSE", "END", "ESCAPE", "EXCEPT", "EXCLUDE", "EXCLUSIVE", "EXISTS", "EXPLAIN", "FAIL", "FILTER", "FIRST", "FOLLOWING", "FOR", "FOREIGN", "FROM", "FULL", "GENERATED", "GLOB", "GROUP", "GROUPS", "HAVING", "IF", "IGNORE", "IMMEDIATE", "IN", "INDEX", "INDEXED", "INITIALLY", "INNER", "INSERT", "INSTEAD", "INTERSECT", "INTO", "IS", "ISNULL", "JOIN", "KEY", "LAST", "LEFT", "LIKE", "LIMIT", "MATCH", "MATERIALIZED", "NATURAL", "NO", "NOT", "NOTHING", "NOTNULL", "NULL", "NULLS", "OF", "OFFSET", "ON", "OR", "ORDER", "OTHERS", "OUTER", "OVER", "PARTITION", "PLAN", "PRAGMA", "PRECEDING", "PRIMARY", "QUERY", "RAISE", "RANGE", "RECURSIVE", "REFERENCES", "REGEXP", "REINDEX", "RELEASE", "RENAME", "REPLACE", "RESTRICT", "RETURNING", "RIGHT", "ROLLBACK", "ROW", "ROWS", "SAVEPOINT", "SELECT", "SET", "TABLE", "TEMP", "TEMPORARY", "THEN", "TIES", "TO", "TRANSACTION", "TRIGGER", "UNBOUNDED", "UNION", "UNIQUE", "UPDATE", "USING", "VACUUM", "VALUES", "VIEW", "VIRTUAL", "WHEN", "WHERE", "WINDOW", "WITH", "WITHOUT" ]
    key_upper = key.upper()
    return key_upper in sqlite_reserved_words

def escape_key_if_required(key):
    if "." in key or is_reserved(key):
        key = f'"{key}"'
    return key

def escaped_keys(input_keys):
    keys = [ escape_key_if_required(key) for key in input_keys ]
    return keys

class SqliteType(IntEnum):
    TEXT = 1
    REAL = 2
    INTEGER = 3

    BLOB = 4 # ideally 2 but confuses from_text
    LIST = 5 # ideally 2 but confuses from_text

    def from_text(arg):
        D = { "TEXT"    : SqliteType.TEXT,    "REAL" : SqliteType.REAL, 
              "INTEGER" : SqliteType.INTEGER, "INT"  : SqliteType.INTEGER, 
              "BLOB"    : SqliteType.BLOB,    "LIST" : SqliteType.LIST }
        return D[arg]

def as_column_str(x):
    D = { str             : "TEXT", float           : "REAL", int                : "INTEGER", bytes          : "BLOB", list            : "LIST", 
          SqliteType.TEXT : "TEXT", SqliteType.REAL : "REAL", SqliteType.INTEGER : "INTEGER", SqliteType.BLOB: "BLOB", SqliteType.LIST : "LIST",
          type(None)      : "TEXT", bool            : "INTEGER" }
    if x not in D.keys():
        x = type(x)
    return D[x]

class Schema:
    "TEXT (null)"
    "TEXT (genuine)"
    "INTEGER"
    "REAL"
    "BLOB"
    "LIST"

    def __init__(self, arg=None):
        if arg == [] or arg == None:
            self.D = {}
        else:
            items = arg if type(arg) == list else [ arg ]
            is_pragma = type(items[0]) == sqlite3.Row
            if is_pragma:
                self.D = { row_D["name"] : SqliteType.from_text(row_D["type"]) for row_D in items }
            else:
                self.D = self.dictionaries_to_schema(items)

    def dictionaries_to_schema(self, dictionaries):
        items = []
        for D in dictionaries:
            for T in D.items():
                items.append(T)

        # timer.start("Implementation 1")
        # D1 = {}
        # content_offset_count = 0
        # for k, v in items:
        #     v1 = D1.get(k, SqliteType.TEXT) # 0.8, v1            = SqliteType.TEXT, INTEGER, REAL
        #     column_str = as_column_str(v)         # 2.3, as_column_str = "TEXT, "INTEGER", "REAL"
        #     v2 = SqliteType.from_text(column_str) # 3.7, v2            = SqliteType.TEXT, INTEGER, REAL
        #     D1[k] = max(v1, v2)             # 0.8, D1[k]   = SqliteType.TEXT, INTEGER, REAL
        #     if k == "content_offset_seconds":
        #         content_offset_count = content_offset_count + 1
        #         # if content_offset_count < 20:
        #         if v2 == 3:
        #             print(f"[{content_offset_count}] content_offset_seconds: {v}, v1: {v1}, v2: {v2}")
        # timer.print("Implementation 1")

        finalized_column_names = []
        type2sqlite = { type(None): SqliteType.TEXT, str: SqliteType.TEXT, float: SqliteType.REAL, int: SqliteType.INTEGER, bool: SqliteType.INTEGER, bytes: SqliteType.BLOB, list: SqliteType.LIST }
        schema_D = {}
        for column_name, value in items:
            if column_name in finalized_column_names:
                continue
            value_type = type(value)
            schema_D[column_name] = type2sqlite[value_type]
            if value != None and value_type != int:
                finalized_column_names.append(column_name)

        return schema_D

    def join(self, with_types=False):
        if not with_types:
            return ", ".join(escaped_keys(self.keys()))
        return ", ".join(escape_key_if_required(key)+" "+as_column_str(value) for key, value in self.items())

    def column_definitions(self, constraint_D={}):
        # https://www.sqlite.org/syntax/column-constraint.html
        column_definitions = []
        for key, value in self.items():
            column_L = [ escape_key_if_required(key), as_column_str(value) ] 
            if constraint_D.get(key):
                column_L = [ escape_key_if_required(key), as_column_str(value), constraint_D.get(key)]
            else:
                column_L = [ escape_key_if_required(key), as_column_str(value) ]
            column_definition = " ".join(column_L)
            column_definitions.append(column_definition)
        return ", ".join(column_definitions)

    def keys(self):
        return self.D.keys()
    def values(self):
        return self.D.values()
    def items(self):
        return self.D.items()
    def get(self, key):
        return self.D.get(key)
    def __getitem__(self, key):
        return self.D[key]
    def __setitem__(self, key, value):
        self.D[key] = value
    def __len__(self):
        return len(self.D)
    def __repr__(self):
        return repr(self.D)
    def __eq__(self, rhs_schema):
        return self.D == rhs_schema.D


    def __str__(self):
        if len(self.D) == 0:
            return "Empty Schema"
        max_length = max(len(key) for key in self.D.keys())
        column_gap = 2
        separator = " " * column_gap
        gutter_length = 2
        gutter_string = " " * gutter_length
        lines = []
        for k, v in self.D.items():
            key_string = str(k).ljust(max_length)
            value_string = str(v)
            lines.append(gutter_string + key_string + separator + value_string)
        lines = [ "" ] + lines + [ "" ]
        return "\n".join(lines)


def get_schema(cursor, table_name):
    schema_query = cursor.execute(f"PRAGMA table_info('{table_name}');")
    rows = cursor.fetchall()
    schema = Schema(rows)
    return schema

def batch_execute(cursor, statements, print_statements=True):
    for statement in statements:
        cursor.execute(statement)
        if print_statements:
            print(statement)
            print()


def add_new_columns(cursor, table_name, dictionary_schema):
    table_schema = get_schema(cursor, table_name)
    new_columns_schema = Schema()
    new_columns_schema.D = { k: v for k, v in dictionary_schema.items() if table_schema.get(k) == None }
    new_column_statements = [ f"ALTER TABLE {table_name} ADD COLUMN {escape_key_if_required(k)} {as_column_str(v)};" for k, v in new_columns_schema.items() ]
    return new_column_statements

def modify_existing_column_types(cursor, table_name, dictionary_schema, constraint_D={}):
    def get_random_four_digit_hex():
        from random import randint
        min_hex = 4096
        max_hex = 65535
        return hex(randint(min_hex, max_hex))[2:]
    # We don’t check if table_schema has a key in dictionary_schema as this is meant to be run AFTER add_new_columns
    # There should be no key in dictionary that is not there in the table schema
    table_schema = get_schema(cursor, table_name)

    changed_columns = [ key for key in dictionary_schema.keys() if table_schema[key] != dictionary_schema[key] ]
    if len(changed_columns) == 0:
        return []

    for key in dictionary_schema.keys():
        table_schema[key] = max(dictionary_schema[key], table_schema[key])

    temp_table_name        = f"tmp_{table_name}_{get_random_four_digit_hex()}"
    create_table_statement = f"CREATE TABLE {temp_table_name} ({table_schema.column_definitions(constraint_D=constraint_D)});"
    key_string             = table_schema.join()
    insert_statement       = f"INSERT INTO {temp_table_name}({key_string}) SELECT {key_string} FROM {table_name};"
    drop_statement         = f"DROP TABLE {table_name};"
    rename_statement       = f"ALTER TABLE {temp_table_name} RENAME TO {table_name};"

    return [ create_table_statement, insert_statement, drop_statement, rename_statement ]

def get_insert_pair_with_on_conflict_clause(D, table_name):
    keys = D.keys()
    values = tuple(D[key] for key in keys)
    keys = escaped_keys(keys)
    column_name_string = ", ".join(keys)
    question_mark_string = ", ".join(["?"] * len(keys))
    prefixed_keys = [ f"excluded.{key}" for key in keys ]
    prefixed_column_name_string = ", ".join(prefixed_keys)
    statement = f"""
INSERT INTO {table_name} ({column_name_string}) VALUES ({question_mark_string})
ON CONFLICT DO UPDATE SET ({column_name_string})=({prefixed_column_name_string});
    """.strip()
    return (statement, values)

# TRIGGER
def get_create_trigger_statement(cursor, table_name, trigger_name, vcs_table_name, date_column_name="vcs_date_epoch_utc_seconds"):
    def get_primary_key(cursor, table_name):
        schema_query = cursor.execute(f"SELECT name FROM pragma_table_info('{table_name}') WHERE pk=1;")
        row = cursor.fetchone()
        primary_key = row["name"]
        return primary_key

    primary_key       = get_primary_key(cursor, table_name)
    table_schema      = get_schema(cursor, table_name)
    unescaped_column_names      = table_schema.keys()
    clauses           = []
    column_names      = []
    column_operations = []
    for unescaped_column_name in unescaped_column_names:
        escaped_column_name = escape_key_if_required(unescaped_column_name)
        old_prefix          = "old."+escaped_column_name
        new_prefix          = "new."+escaped_column_name
        condition           = f"{old_prefix} IS NOT {new_prefix}"
        column_operation    = f"NULLIF({old_prefix}, {new_prefix})"
        if unescaped_column_name != primary_key and unescaped_column_name != date_column_name:
            clauses.append(condition)
            column_names.append(escaped_column_name)
            column_operations.append(column_operation)
        else:
            column_names.append(escaped_column_name)
            column_operations.append(old_prefix)
    if len(clauses) == 0:
        return ""
    condition_string = " OR ".join(clauses)
    column_name_string = ", ".join(column_names)
    column_operation_string = ", ".join(column_operations)
    create_trigger_statement = f"""
CREATE TRIGGER {trigger_name}
    BEFORE UPDATE
    ON {table_name}
    WHEN {condition_string}
BEGIN
    INSERT INTO {vcs_table_name} ({column_name_string}) VALUES (
        {column_operation_string}
    );
END;
    """.strip()
    if column_name_string == "":
        print_red_line("column_name_string is EMPTY")
        breakpoint()
    return create_trigger_statement

# TRIGGER
def check_if_trigger_exists(cursor, table_name, trigger_name):
    query = f"SELECT COUNT(*) trigger_count FROM sqlite_master WHERE type='trigger' AND tbl_name='{table_name}' AND name='{trigger_name}';"
    row = cursor.execute(query).fetchone()
    trigger_count = row["trigger_count"]
    return trigger_count > 0

# TRIGGER
def add_date_column(dictionaries):
    from time import time
    epoch_utc = int(time())
    date_column_name = "vcs_date_epoch_utc_seconds"
    for D in dictionaries:
        D[date_column_name] = epoch_utc
    return dictionaries

def insert_dictionaries(cursor, table_name, dictionaries, constraint_D={}, vcs=False):
    if dictionaries == []:
        return

    import threading
    global lock
    lock = threading.Lock()
    lock.acquire()
    print_statements = False

    dictionaries = [ prepare_dict(D) for D in dictionaries ]
    if vcs:
        dictionaries = add_date_column(dictionaries)

    dictionary_schema = Schema(dictionaries)

    table_schema = get_schema(cursor, table_name)
    table_does_not_exist = len(table_schema) == 0

    # 1. CREATE TABLE IF IT DOES NOT EXIST
    if table_does_not_exist:
        create_table_statement = f"CREATE TABLE {table_name} ({dictionary_schema.column_definitions(constraint_D=constraint_D)});"
        cursor.execute(create_table_statement)
        if print_statements:
            print(create_table_statement)

    # 2. ADD NEW COLUMNS
    add_column_statements = add_new_columns(cursor, table_name, dictionary_schema)
    batch_execute(cursor, add_column_statements, print_statements)

    # 3. MODIFY UPDATED COLUMN TYPES
    modify_column_statements = modify_existing_column_types(cursor, table_name, dictionary_schema, constraint_D=constraint_D)
    batch_execute(cursor, modify_column_statements, print_statements)

    if vcs:
        # 4.1. SETUP VCS TABLE {{{
        vcs_print_statements = False
        vcs_table_name = f"{table_name}_vcs"
        vcs_table_schema = get_schema(cursor, vcs_table_name)
        vcs_table_does_not_exist = len(vcs_table_schema) == 0

        if vcs_table_does_not_exist:
            create_vcs_table_statement = f"CREATE TABLE {vcs_table_name} ({dictionary_schema.column_definitions()});"
            cursor.execute(create_vcs_table_statement)
            if vcs_print_statements:
                print(create_vcs_table_statement)
        vcs_add_column_statements = add_new_columns(cursor, vcs_table_name, dictionary_schema)
        batch_execute(cursor, vcs_add_column_statements, vcs_print_statements)
        vcs_modify_column_statements = modify_existing_column_types(cursor, vcs_table_name, dictionary_schema)
        batch_execute(cursor, vcs_modify_column_statements, vcs_print_statements)
        # }}}
        # 4.2. ADD VCS TRIGGER
        trigger_name = f"update_trigger_{table_name}_vcs"
        if len(vcs_add_column_statements) > 0 or len(vcs_modify_column_statements) > 0: # SNIPPET 1 for debug
            cursor.execute(f"DROP TRIGGER IF EXISTS {trigger_name}")
            if vcs_print_statements:
                print(f"DROP TRIGGER IF EXISTS {trigger_name}")
        trigger_exists = check_if_trigger_exists(cursor, table_name, trigger_name)
        if not trigger_exists:
            primary_key_supplied = len([ value for value in constraint_D.values() if value.index("PRIMARY KEY") > -1 ]) == 1
            if primary_key_supplied:
                create_trigger_statement = get_create_trigger_statement(cursor, table_name, trigger_name, vcs_table_name)
                cursor.execute(create_trigger_statement)
                if vcs_print_statements:
                    print(create_trigger_statement)

    # 5. INSERT
    for D in dictionaries:
        insert_statement, values = get_insert_pair_with_on_conflict_clause(D, table_name)
        try:
            cursor.execute(insert_statement, values )
        except Exception as e:
            print()
            print("[ERROR]", insert_statement)
            raise
        if print_statements:
            print(insert_statement, make_string_green(truncate(str(values), 200)))
            print()
    lock.release()

def version():
    pysqlite_version = sqlite3.version
    sqlite_version = sqlite3.sqlite_version
    print(f"PYSQLITE: {pysqlite_version}")
    print(f"SQLITE  : {sqlite_version}")

def main():
    from tdd import test_ic_custom_types
    test_ic_custom_types()
    pass
    # list_type()

if __name__ == "__main__":
    # import sys; main(); sys.exit(1)
    from .error import handler
    with handler():
        main()




# import sys; sys.path = [ "C:/users/vivek/Documents/Python/" ] + sys.path
from .timer import Timer   # Single Use
timer = Timer()
from .color import Code    # Multi-Use
from .error import handler # Single Use
from .ic import ic, ib     # Multi-Use, import time: 70ms - 110ms
Code; ic; ib; handler
from .reltools import rel2abs

def main():
    # print_python_modules()
    # print()
    # print_pipreqs()
    pypi_version()

def pypi_version():
    from .ic import ic
    import requests
    package_name = "waivek"
    url = f"https://pypi.org/pypi/{package_name}/json"
    resp = requests.get(url)
    D = resp.json()
    version = D["info"]["version"]
    major, minor, patch = version.split(".")
    patch = int(patch) + 1
    new_version = f"V- {major}.{minor}.{patch}"
    ic(new_version)
    


def print_python_modules():
    import os.path
    to_module_name = lambda path: os.path.splitext(os.path.basename(path))[0]
    helper_modules = [ 'frame.py', 'trace.py' ]
    ignore_modules = [ 'reqs.py', 'setup.py', 'tdd.py', 'template.py', 'test.py' ]
    py_modules = [ to_module_name(module) for module in get_python_files() if module not in ignore_modules ]
    print(f"py_modules = {py_modules}")
    print("# [print_python_modules] see helper_modules and ignore_modules for ignored python files")

def print_pipreqs():
    import subprocess
    current_dir = rel2abs(".")
    command = ["pipreqs", "--print", current_dir ]
    print("$ " + " ".join(command))
    result_binary = subprocess.run(command, capture_output=True)
    contents = result_binary.stdout.decode("utf-8").strip()
    lines = contents.split("\n")
    table = []
    manual_lines = [
        'pysqlite3-binary; platform_system=="Linux"'
    ]
    required_packages = {
        "aiohttp": "get.py",
        "executing": "ic.py",
        "python_dateutil": "common.py",
        "timeago": "common.py"
    }
    ignore_packages = {
        "ipdb": "error.py",
        "pysqlite3": "db.py"
    }
    for line in lines:
        package, version = line.split("==")
        table.append({ "package": package, "version": version })

    install_requires_lines = []
    for D in table:
        package = D['package']
        version = D['version']
        if package in required_packages:
            print(Code.GREEN + package)
            line = f"{package}=={version}"
            install_requires_lines.append(line)
        elif package in ignore_packages:
            print(Code.LIGHTBLACK_EX + (package + "# ignore"))
        else:
            print(Code.RED + package)
    lines = install_requires_lines + manual_lines
    lines = [ line.strip() for line in lines ]
    lines = [ f"'{line}'" for line in lines ]
    lines = [ f"        {line}" for line in lines ]
    print("    install_requires = [")
    print(",\n".join(lines))
    print("    ]")


            
        


def get_directories():
    from glob import glob
    import os.path
    current_dir = rel2abs(".")
    paths = glob(os.path.join(current_dir, "**/"), recursive=True)
    relpaths = [ os.path.relpath(path, current_dir) for path in paths ]
    return relpaths

def get_non_python_files():
    from glob import glob
    import os.path
    current_dir = rel2abs(".")
    paths = glob(os.path.join(current_dir, "*"), recursive=False)
    paths = [ path for path in paths if os.path.isfile(path) ]
    relpaths = [ os.path.relpath(path, current_dir) for path in paths ]
    relpaths = [ path for path in relpaths if not path.endswith(".py") ]
    return relpaths

def get_python_files():
    from glob import glob
    import os.path
    current_dir = rel2abs(".")
    paths = glob(os.path.join(current_dir, "**/*.py"), recursive=True)
    relpaths = [ os.path.relpath(path, current_dir) for path in paths ]
    return relpaths

def get_tracked_python_files():
    import git
    repo = git.Repo(".")
    tracked_files = repo.git.ls_files().split("\n")
    tracked_files = [ file for file in tracked_files if file.endswith(".py") ]
    return tracked_files

def get_untracked_python_files():
    import git
    repo = git.Repo(".")
    untracked_files = repo.untracked_files
    untracked_files = [ file for file in untracked_files if file.endswith(".py") ]
    return untracked_files

if __name__ == "__main__":
    with handler():
        main()




# from colorama import init, Fore
# init(convert=True)
from .color import Code

# print_object {{{
def print_object(obj, hidden=False):
    if not hidden:
        keys = [ key for key in dir(obj) if not key.startswith("__") ]
    else:
        keys = dir(obj)

    D = { key : obj.__getattribute__(key) for key in keys }
    print_dict(D)
    # }}}

# print_dict {{{
def print_dict(D):
    from columnar import columnar
    if type(D) != type({}):
        argument_type = str(type(D))
        str_representation = truncate(str(D), 40)
        print_red_line("Please pass a dictionary. Invalid type {arg_type} with str representation {str_rep}".format(
            arg_type=argument_type, str_rep=str_representation))
        return
    data = []
    for key, value in D.items():
        trunc_val = truncate(str(value), 80)
        data.append([key, trunc_val])
    headers = [ "key", "value" ]
    table = columnar(data, headers, no_borders=True)
    print(table)
    # }}}

# print_table {{{
# Assumption: Columns have consistent type
def print_table(dictionaries):
    if len(dictionaries) == 0:
        return
    for D in dictionaries:
        for key in D.keys():
            value = D[key]
            if type(value) != str and type(value) != int and type(value) != float:
                D[key] = str(value)
    # keys = ["basename", "file_read_time", "json_load_time", "total_time"]
    keys = list(dictionaries[0].keys())
    # keys = ["name", "cost", "count"]
    max_length_D = {key:len(key) for key in keys}
    types_D = { key: type(dictionaries[0][key]) for key in keys }
    left_gutter_length = 2
    vertical_padding = 1
    left_gutter_string = " " * left_gutter_length
    column_gap = 2
    separator = " "  * column_gap
    for key in keys:
        max_length = max_length_D[key]
        for D in dictionaries:
            value = D[key]
            if type(value) == str:
                length = len(value)
            if type(value) == int:
                length = len(str(value))
            if type(value) == float:
                length = len("%.2f" % value)
            if max_length < length:
                max_length = length
        max_length_D[key] = max_length

    for i in range(vertical_padding):
        print()
    column_label_items = []
    for key in keys:
        if types_D[key] == str:
            rep = key.ljust(max_length_D[key])
        if types_D[key] == int:
            rep = key.rjust(max_length_D[key])
        if types_D[key] == float:
            rep = key.rjust(max_length_D[key])
        column_label_items.append(rep)
    line = left_gutter_string + separator.join(column_label_items)
    print(line)


    for D in dictionaries:
        line_items = []
        for key in keys:
            max_length = max_length_D[key]
            value = D[key]
            if type(value) == str:
                rep = value.ljust(max_length)
            if type(value) == int:
                rep = str(value).rjust(max_length)
            if type(value) == float:
                rep = ("%.2f" % value).rjust(max_length)
            line_items.append(rep)
        line = left_gutter_string + separator.join(line_items)
        print(line)
    for i in range(vertical_padding):
        print()
# }}}

# head {{{
def head_file(file_contents, preview_lines_count=5):
    lines = file_contents.split("\n")
    preview_lines = lines[0:preview_lines_count]
    return "\n".join(preview_lines)

def head_list(L, preview_lines_count=5):
    preview_L = L[0:preview_lines_count]
    return "\n".join(preview_L)

def head(inp, preview_lines_count=5):
    if type(inp) == list:
        return head_list(inp, preview_lines_count)
    else:
        return head_file(inp, preview_lines_count)
# }}}

# print_red_line, print_green_line {{{
def make_string_green(string):
    # return "{color_code}{string}{reset_code}".format(color_code=Fore.GREEN, string=string, reset_code=Fore.RESET)
    return Code.GREEN + string
def make_string_red(string):
    # return "{color_code}{string}{reset_code}".format(color_code=Fore.RED, string=string, reset_code=Fore.RESET)
    return Code.RED + string
def print_red_line(string):
    print(make_string_red(string))
def print_green_line(string):
    print(make_string_green(string))
    # }}}

# truncate {{{
def truncate(string, length):
    if len(string) < length:
        return string

    middle_string = "..."
    slice_size = int((length-len(middle_string))/2)
    # 80, 38, 38, 3 SUM = 79/80 ONE SPACE AVAILABLE
    # 81, 39, 39, 3 SUM = 81/81 ZERO SPACES AVAILABLE
    even_string_length = length % 2 == 0
    left_end = slice_size
    if even_string_length:
        right_start = slice_size+1
    else:
        right_start = slice_size
    left_slice = string[0:left_end]
    right_slice = string[-right_start:]
    trunc_string = "{left_slice}{middle_string}{right_slice}".format(left_slice=left_slice, middle_string=middle_string, right_slice=right_slice)
    if len(trunc_string) != length:
        print_red_line("ERROR: Could not return a truncated string")
        return "ERROR"
    return trunc_string
# }}}

# abbreviate {{{
def abbreviate(number):
    number_int = int(number)
    number_string = str(number_int)
    number_length = len(str(number_int))
    if number_int < 100:
        return number_string
    first_digit, second_digit, third_digit = number_string[0:3]
    abbreviation_dict = {
         1:  "",  2:  "",  3:  "",
         4: "k",  5: "k",  6: "k",
         7: "m",  8: "m",  9: "m",
        10: "b", 11: "b", 12: "b",
        13: "t", 14: "t", 15: "t"
    }
    letter = abbreviation_dict[number_length]
    if number_length % 3 == 0:
        # first three digits
        abbr_string = f"{first_digit}{second_digit}{third_digit}{letter}"
    if number_length % 3 == 2:
        # first two digits, period, next digit
        abbr_string = f"{first_digit}{second_digit}.{third_digit}{letter}"
    if number_length % 3 == 1:
        # first digit, period, next two digits
        abbr_string = f"{first_digit}.{second_digit}{third_digit}{letter}"

    return abbr_string
# }}}




from .color import Code
# from .ic import ic, ib

from contextlib import contextmanager

def stub_quiet():
    from .ic import ic, ib
    ic
    ib

def get_args(frame, summary):
    import inspect # builtin
    if summary.name in [ "<module>", "<listcomp>" ]:
        # Line is not inside a function
        return []
    if summary.name in frame.f_globals.keys():
        # Line is inside a GLOBAL function
        return inspect.getfullargspec(frame.f_globals[summary.name]).args
    # Line is inside a LOCAL function of a class
    import gc # builtin
    # https://stackoverflow.com/a/52762678
    try:
        function_object = [obj for obj in gc.get_referrers(frame.f_code) if hasattr(obj, '__code__') and obj.__code__ is frame.f_code][0]
    except Exception as e:
        error = e
    return inspect.getfullargspec(function_object).args

class Frame:
    def __init__(self, frame, summary):
        from os.path import basename
        keys = get_args(frame, summary)
        self.signature = "(" + ", ".join(keys) + ")"
        self.method = summary.name
        self.locals = frame.f_locals 
        self.line_number = str(summary.lineno)
        self.path = basename(summary.filename)

    def pretty(self):
        from .ic import Table
        header = "\n".join([ self.path+":"+self.line_number, self.method + self.signature ])

        rows = [ (k, v) for k, v in self.locals.items() ]
        table = Table()
        for row in rows:
            table.row(row)

        import textwrap # built-in
        if table.table:
            table_string = "\n".join(str(table).split("\n")[1:])
        else:
            table_string = "    (empty table)"
        table_string = textwrap.dedent(table_string)
        table_string = textwrap.indent(table_string, "    ")
        # print(header + "\n" + table_string)
        print(header)
        print_variables(self.locals)

    def __repr__(self):
        repr_string = f"Frame(method={self.method}, locals={self.locals})"
        return repr_string

    def __getitem__(self, key):
        return self.locals[key]
    def __str__(self):
        return repr(self)


class Frames:
    def __init__(self, error):
        import traceback
        tb = error.__traceback__
        frames = [ frame for frame, _ in traceback.walk_tb(tb) ]
        summaries = traceback.extract_tb(tb)
        pairs = list(reversed(list(zip(frames, summaries))))
        self.frames = []
        for frame, summary in pairs:

            frm = Frame(frame, summary)
            self.frames.append(frm)
        # self.frames = [ Frame(frame, summary) for frame, summary in pairs ]

    def __getitem__(self, index):
        return self.frames[index]

    def __str__(self):
        return str(self.frames)

    def __repr__(self):
        return str(self.frames)

    def pretty(self):
        from .ic import Table
        from .common import truncate
        dictionaries = [ ( "i", "method", "local_count", "locals" ) ]
        dictionaries = dictionaries + [ ( i, D.method, D.local_count, truncate(str(D.locals), 120) ) for i, D in enumerate(self.frames) ]
        table = Table()
        for row in dictionaries:
            table.row(row)

    def print_stack(self):
        print()
        for frame in self.frames:
            frame.pretty()

def color_D_if_big(D):
    if len(str(D)) < 120:
        return str(D)
    else:
        colored_items = []
        for i, (k, v) in enumerate(D.items()):
            code = Code.CYAN if i % 2 == 0 else Code.LIGHTCYAN_EX
            # code = "[cyan]" if i % 2 == 0 else "[blue]"
            colored_items.append(code + f"{repr(k)}: {repr(v)}")
        return "{" + ", ".join(colored_items) + "}"

def print_variables_by_frame(error):
    import traceback
    from os.path import basename
    import os
    import textwrap
    import types
    from .ic import Table

    tb = error.__traceback__
    frames = [ frame for frame, _ in traceback.walk_tb(tb) ]
    summaries = traceback.extract_tb(tb)
    pairs = list(reversed(list(zip(frames, summaries))))
    exclude_types = [types.FunctionType, types.ModuleType, types.BuiltinFunctionType, type]
    tuples = []
    table = Table()

    method_lines = []
    for frame, summary in pairs:
        try:
            keys = get_args(frame, summary)
        except Exception as e:
            error = e
            print(Code.RED + repr(error))
        signature = "(" + ", ".join(keys) + ")"
        method = summary.name
        method_line = method + signature if summary.name != "<module>" else "<module>"
        method_lines.append(method_line)
    METHOD_LINE_LENGTH_LIMIT = 60
    variable_line_width = os.get_terminal_size()[0] - 2*len(table.gutter) - len(table.separator) - min(max(len(method_line) for method_line in method_lines), METHOD_LINE_LENGTH_LIMIT)
    for frame, summary in pairs:
        keys = get_args(frame, summary)
        signature = "(" + ", ".join(keys) + ")"
        method = summary.name
        local_D = frame.f_locals 
        method_line = method + signature if summary.name != "<module>" else "<module>"
        path_line = basename(summary.filename) + ":" + str(summary.lineno)
        variable_D = { k:v for k, v in local_D.items() 
                if type(v) not in exclude_types and k[0:2] != "__" and str(v)[0] != "<"  and not isinstance(v, Exception) }
        variable_lines = textwrap.wrap(color_D_if_big(variable_D), variable_line_width)
        variable_line = "\n".join(variable_lines)
        # print(method_line)
        # print(path_line)
        method_line_wrapped = "\n".join(textwrap.wrap(method_line, METHOD_LINE_LENGTH_LIMIT))
        table.row([method_line_wrapped, variable_line])
    table_string = textwrap.indent(textwrap.dedent("\n".join(str(table).split("\n")[1:])), '    ')
    print(Code.LIGHTCYAN_EX + 'Variables by frame')
    print(table_string)

def color_error_repr_old(error):
    string = repr(error)
    lhs, rhs = string.split("'", 1)
    rhs = "'" + rhs[:-1]
    lhs = Code.LIGHTRED_EX + lhs 
    rhs = (Code.YELLOW + rhs) + (Code.LIGHTRED_EX + ')')
    colored_string = lhs + rhs
    return colored_string

def color_error_regex(error):
    string = repr(error)
    import re
    m = re.match(r"""(\w+\(["'])(.*)(["']\))""", string)
    
    lhs = Code.LIGHTRED_EX + m.group(1) 
    middle =  Code.YELLOW + m.group(2) 
    # middle = re.sub(r"('\w+')", Code.LIGHTCYAN_EX + r'\1', middle)
    rhs = Code.LIGHTRED_EX + m.group(3)
    colored_string = lhs + middle + rhs
    return colored_string

def color_arg(arg):
    BLACK           = '\x1b[30m'
    RED             = '\x1b[31m'
    GREEN           = '\x1b[32m'
    YELLOW          = '\x1b[33m'
    BLUE            = '\x1b[34m'
    MAGENTA         = '\x1b[35m'
    CYAN            = '\x1b[36m'
    WHITE           = '\x1b[37m'
    LIGHTBLACK_EX   = '\x1b[90m'
    LIGHTRED_EX     = '\x1b[91m'
    LIGHTGREEN_EX   = '\x1b[92m'
    LIGHTYELLOW_EX  = '\x1b[93m'
    LIGHTBLUE_EX    = '\x1b[94m'
    LIGHTMAGENTA_EX = '\x1b[95m'
    LIGHTCYAN_EX    = '\x1b[96m'
    LIGHTWHITE_EX   = '\x1b[97m'
    RESET = '\x1b[39m'

    if type(arg) == str:
        arg = YELLOW + repr(arg)
    else:
        arg = CYAN + repr(arg)
    return arg


def color_error_repr(error):
    BLACK           = '\x1b[30m'
    RED             = '\x1b[31m'
    GREEN           = '\x1b[32m'
    YELLOW          = '\x1b[33m'
    BLUE            = '\x1b[34m'
    MAGENTA         = '\x1b[35m'
    CYAN            = '\x1b[36m'
    WHITE           = '\x1b[37m'
    LIGHTBLACK_EX   = '\x1b[90m'
    LIGHTRED_EX     = '\x1b[91m'
    LIGHTGREEN_EX   = '\x1b[92m'
    LIGHTYELLOW_EX  = '\x1b[93m'
    LIGHTBLUE_EX    = '\x1b[94m'
    LIGHTMAGENTA_EX = '\x1b[95m'
    LIGHTCYAN_EX    = '\x1b[96m'
    LIGHTWHITE_EX   = '\x1b[97m'
    RESET = '\x1b[39m'

    string = repr(error)
    args = ', '.join(color_arg(arg) for arg in error.args)
    colored_string = f"{LIGHTRED_EX}{error.__class__.__name__}({YELLOW}{args}{LIGHTRED_EX}){RESET}"
    return colored_string
    # lhs = Code.LIGHTRED_EX + m.group(1) 
    # middle =  Code.YELLOW + m.group(2) 
    # # middle = re.sub(r"('\w+')", Code.LIGHTCYAN_EX + r'\1', middle)
    # rhs = Code.LIGHTRED_EX + m.group(3)
    # colored_string = lhs + middle + rhs
    return colored_string


def print_error_information(error):
    # C:\Users\vivek\Documents\Python                               -> ~/Documents/Python
    # C:\Users\vivek\AppData\Roaming\Python\Python310\site-packages -> $APPDATA/Python/site-packages
    # C:\Program Files\Python310\Lib\site-packages                  -> 
    import traceback
    import os
    from pathlib import Path

    tb = error.__traceback__

    print()
    print(color_error_repr(error))
    # print(Code.LIGHTRED_EX + repr(error))

    from .frame import frame_gen
    from .ic import ic; ic
    call_frames = list(frame_gen())
    call_file = call_frames[3].f_code.co_filename

    frames = [ frame for frame, _ in traceback.walk_tb(tb) ]

    summaries = traceback.extract_tb(tb)
    pairs = reversed(list(zip(frames, summaries)))
    
    from .ic import Table
    table = Table()
    table.gutter = '    '
    table.separator = ' ... '

    green_done = False
    for i, (frame, summary) in enumerate(pairs):
        filepath = summary.filename
        filepath = str(Path(filepath).resolve()) # C:\users -> C:\Users
        if os.getcwd() in filepath:
            filepath = os.path.relpath(os.path.abspath(filepath))
        else:
            homedir = os.path.expanduser("~")
            filepath = filepath.replace(homedir, "~")

        line_number = summary.lineno
        line = summary.line
        # line = Code.LIGHTGREEN_EX + line if i == 0 else line
        if green_done is False and frame.f_code.co_filename == call_file:
            line = Code.LIGHTGREEN_EX + line
            green_done = True
        lhs_string = f"{filepath}:{line_number}"
        table.row([lhs_string, line])
    table_string = str(table)
    table_lines = table_string.split("\n")
    table_lines = table_lines[1:]
    table_string = "\n".join(table_lines)
    print(table_string)
    # print()

def print_variables(D):
    import types
    from .common import truncate
    from .ic import ic
    exclude_types = [types.FunctionType, types.ModuleType, types.BuiltinFunctionType, type]
    table = [ {"name":k,"type":type(v), "value": truncate(str(v), 160)} for k,v in D.items() 
              if type(v) not in exclude_types and k[0:2] != "__" and str(v)[0] != "<"  and not isinstance(v, Exception) ]
    ic(table)

def get_error_filepath():
    import platform
    import os.path
    if platform.system() == 'Linux':
        vimpath = "~/.vim"
    else:
        vimpath = "~/.vimfiles"
    folder = os.path.join(os.path.expanduser(vimpath), "temp")
    os.makedirs(folder, exist_ok=True)
    filename = 'handler_error.txt'
    filepath = os.path.join(folder, filename)
    filepath = os.path.abspath(filepath)
    return filepath



def write_vim_error_file(error: Exception):
    # test.py:34:5:    import sys
    import linecache
    import traceback
    import re
    from .ic import ic; ic
    import os

    from .frame import frame_gen
    call_frames = list(frame_gen())
    call_file = call_frames[3].f_code.co_filename

    filepath = get_error_filepath()
    frames = [ frame for frame, _ in reversed(list(traceback.walk_tb(error.__traceback__))) ]
    error_lines = []
    cc_nr = -1
    for index, frame in enumerate(frames):
        path = frame.f_code.co_filename
        lineno = frame.f_lineno
        line = linecache.getline(path, lineno)
        if cc_nr == -1 and path == call_file:
            cc_nr = index + 1
        column = re.search(r"\S", line).start() + 1
        error_line = f"{path}:{lineno}:{column}:{line}".rstrip()
        error_lines.append(error_line)
    with open(filepath, "w") as f:
        f.write("\n".join(error_lines))
    # we use threading to make os\.system call not take up 0.08 seconds
    gvim_command = lambda : os.system(f'''start gvim --servername GVIM --remote-send ":call RHEF('{filepath}', {cc_nr})<CR>" ''')
    import threading
    threading.Thread(target=gvim_command).start()


@contextmanager
def handler():
    try:
        yield
    except Exception as e:
        error = e
        if type(e).__name__ == 'bdb.BdbQuit':
            # Exit Via CTRL-D
            pass
        else:
            write_vim_error_file(e)
            print_error_information(e)
            # print_variables_by_frame(e)
            import sys
            # frames = Frames(error)

            # for frame in frames:
            #
            #     print(frame.method)

            print(Code.LIGHTCYAN_EX + "alias ic, ib | import sys, ic, ib")
            print()
            # try __builtins__.__dict__
            # if "ipython" in sys.argv[0] or __builtins__.get("get_ipython", False):
            if "ipython" in sys.argv[0]:
                import ipdb
                ipdb.post_mortem(e.__traceback__)
            else:
                import pdb
                pdb.post_mortem(e.__traceback__)

def divide_by_zero():

    # Error 1:
    # ========
    # upper = 5
    # lower = 0
    # result = upper / lower

    # Error 2:
    # ========
    # item = bytes()
    # bytes.encode()


    # Error 3:
    # ========
    from .reltools import here
    path = here() / "f1/f2/f3/item.txt"
    path.mkdir(exist_ok=True)

    result = 1
    return result

def table_wide():
    pass
    # from .ic import Table
    # import textwrap
    # import json
    # # L = [['_make_api_call(self, operation_name, api_params)', '{\'operation_name\': \'PutIntegration\',\'api_params\': {\'restApiId\': \'p343uq88aa\', \'resourceId\':\'tbfu9qe6w7\', \'type\': \'HTTP_PROXY\', \'httpMethod\': \'ANY\',\'integrationHttpMethod\': \'ANY\', \'uri\': \'gql.twitch.tv\',\'connectionType\': \'INTERNET\', \'requestParameters\':{\'integration.request.path.proxy\':\'method.request.path.proxy\',\'integration.request.header.X-Forwarded-For\':\'method.request.header.X-My-X-Forwarded-For\'}},\'operation_model\':OperationModel(name=PutIntegration),\'service_name\': \'apigateway\',\'request_context\': {\'client_region\': \'us-east-2\',\'client_config\': <botocore.config.Config object at0x00000241D3DE7730>, \'has_streaming_input\': False,\'auth_type\': None, \'timestamp\': \'20220209T204914Z\'},\'request_dict\': {\'url_path\': \'/restapis/p343uq88aa/resources/tbfu9qe6w7/methods/ANY/integration\', \'query_string\':{}, \'method\': \'PUT\', \'headers\': {\'Content-Type\':\'application/json\', \'User-Agent\': \'Boto3/1.20.37Python/3.10.1 Windows/10 Botocore/1.23.37\', \'Accept\':\'application/json\'}, \'body\': b\'{"type": "HTTP_PROXY","httpMethod": "ANY", "uri": "gql.twitch.tv","connectionType": "INTERNET", "requestParameters":{"integration.request.path.proxy":"method.request.path.proxy","integration.request.header.X-Forwarded-For":"method.request.header.X-My-X-Forwarded-For"}}\', \'url\':\'https://apigateway.us-east-2.amazonaws.com/restapis/p343uq88aa/resources/tbfu9qe6w7/methods/ANY/integration\', \'context\': {\'client_region\': \'us-east-2\', \'client_config\': <botocore.config.Config object at0x00000241D3DE7730>, \'has_streaming_input\': False,\'auth_type\': None, \'timestamp\': \'20220209T204914Z\'}},\'service_id\': \'api-gateway\',\'event_response\': None, \'parsed_response\':{\'Error\': {\'Message\': \'Invalid HTTP endpoint specified forURI\', \'Code\': \'BadRequestException\'}, \'ResponseMetadata\':{\'RequestId\': \'31570f5f-1014-4d12-a555-9ba1b845a0d9\',\'HTTPStatusCode\': 400, \'HTTPHeaders\': {\'date\': \'Wed, 09 Feb2022 20:49:14 GMT\', \'content-type\': \'application/json\',\'content-length\': \'54\', \'connection\': \'keep-alive\', \'x-amzn-requestid\': \'31570f5f-1014-4d12-a555-9ba1b845a0d9\', \'x-amzn-errortype\': \'BadRequestException\', \'x-amz-apigw-id\':\'NStPrLmBCYcEdtQ=\'}, \'RetryAttempts\': 0}, \'message\':\'Invalid HTTP endpoint specified for URI\'},\'error_code\': \'BadRequestException\'}'], ['_api_call(self)', "{'args': (), 'kwargs': {'restApiId':'p343uq88aa', 'resourceId': 'tbfu9qe6w7', 'type':'HTTP_PROXY', 'httpMethod': 'ANY', 'integrationHttpMethod':'ANY', 'uri': 'gql.twitch.tv', 'connectionType': 'INTERNET','requestParameters': {'integration.request.path.proxy':'method.request.path.proxy','integration.request.header.X-Forwarded-For':'method.request.header.X-My-X-Forwarded-For'}},'operation_name': 'PutIntegration','py_operation_name': 'put_integration'}"], ['init_gateway(self, region, force)', "{'region': 'us-east-2', 'force': False,'session': Session(region_name='us-west-2'),'current_apis': [{'id': 'dtob7epyn2', 'name':'https://api.ipify.org/ - IP Rotate API', 'createdDate':datetime.datetime(2022, 2, 10, 1, 5, 41, tzinfo=tzlocal()),'apiKeySource': 'HEADER', 'endpointConfiguration': {'types':['REGIONAL']}, 'disableExecuteApiEndpoint': False}, {'id':'hfzgbt7mo0', 'name': 'https://gql.twitch.tv/ - IP RotateAPI', 'createdDate': datetime.datetime(2022, 1, 15, 18, 31,15, tzinfo=tzlocal()), 'apiKeySource': 'HEADER','endpointConfiguration': {'types': ['REGIONAL']},'disableExecuteApiEndpoint': False}], 'api':{'id': 'hfzgbt7mo0', 'name': 'https://gql.twitch.tv/ - IPRotate API', 'createdDate': datetime.datetime(2022, 1, 15,18, 31, 15, tzinfo=tzlocal()), 'apiKeySource': 'HEADER','endpointConfiguration': {'types': ['REGIONAL']},'disableExecuteApiEndpoint': False},'create_api_response': {'ResponseMetadata':{'RequestId': 'b46c081f-8baa-4ee4-99a8-5c76ef0a419b','HTTPStatusCode': 201, 'HTTPHeaders': {'date': 'Wed, 09 Feb2022 20:49:13 GMT', 'content-type': 'application/json','content-length': '202', 'connection': 'keep-alive','x-amzn-requestid': 'b46c081f-8baa-4ee4-99a8-5c76ef0a419b','x-amz-apigw-id': 'NStPcLEoiYcEb-Q='}, 'RetryAttempts': 0},'id': 'p343uq88aa', 'name': 'gql.twitch.tv - IP Rotate API','createdDate': datetime.datetime(2022, 2, 10, 2, 19, 13,tzinfo=tzlocal()), 'apiKeySource': 'HEADER','endpointConfiguration': {'types': ['REGIONAL']},'disableExecuteApiEndpoint': False},'get_resource_response': {'ResponseMetadata':{'RequestId': 'd7a96d82-3eca-4b6c-b602-e9944b56e009','HTTPStatusCode': 200, 'HTTPHeaders': {'date': 'Wed, 09 Feb2022 20:49:13 GMT', 'content-type': 'application/json','content-length': '42', 'connection': 'keep-alive', 'x-amzn-requestid': 'd7a96d82-3eca-4b6c-b602-e9944b56e009', 'x-amz-apigw-id': 'NStPiKaLCYcEbpA='}, 'RetryAttempts': 0},'items': [{'id': 'tbfu9qe6w7', 'path': '/'}]},'rest_api_id': 'p343uq88aa','create_resource_response': {'ResponseMetadata':{'RequestId': '61d46688-1caf-4e87-8e11-d564d4a6904d','HTTPStatusCode': 201, 'HTTPHeaders': {'date': 'Wed, 09 Feb2022 20:49:13 GMT', 'content-type': 'application/json','content-length': '81', 'connection': 'keep-alive', 'x-amzn-requestid': '61d46688-1caf-4e87-8e11-d564d4a6904d', 'x-amz-apigw-id': 'NStPlLrMiYcEd1w='}, 'RetryAttempts': 0}, 'id':'830kkp', 'parentId': 'tbfu9qe6w7', 'pathPart': '{proxy+}','path': '/{proxy+}'}}"], ['run(self)', "{'self': None}"], ['__get_result(self)', "{'self': None}"], ['result(self, timeout)', "{'self': None, 'timeout': None}"], ['start(self, force, endpoints)', "{'force': False, 'endpoints': [],'new_endpoints': 0, 'futures': [<Future at0x241d10a49d0 state=finished returned dict>, <Future at0x241d10dc970 state=finished raised BadRequestException>,<Future at 0x241d10f8af0 state=finished returned dict>,<Future at 0x241d1128c10 state=finished returned dict>,<Future at 0x241d1128f40 state=finished returned dict>,<Future at 0x241d116cd00 state=finished returned dict>,<Future at 0x241d118cd30 state=finished returned dict>,<Future at 0x241d11b8d60 state=finished returned dict>,<Future at 0x241d11d8d90 state=finished returned dict>,<Future at 0x241d12228c0 state=finished returneddict>], 'region': 'ca-central-1','result': {'success': True, 'endpoint':'mpwkk460o2.execute-api.us-west-2.amazonaws.com', 'new':False}}"], ['post_anonymous(url, data, json, params, headers, cookies, files, auth, timeout, allow_redirects, proxies, hooks, stream, verify, cert)', '{\'url\': \'https://gql.twitch.tv/gql\', \'data\':None, \'json\': {\'query\': \' query {c0:clip(slug:"AbrasiveGeniusClintTheRinger-eEswgwJ6b0pSfSEC"){slug, videoOffsetSeconds}c1:clip(slug:"HappyArborealClipsmomBlargNaut-Qd9yKSJdoj6ghj1V"){slug, videoOffsetSeconds}c2:clip(slug:"SpineyMuddyPuddingYouDontSay-qYSaw1qvv4bbn_D5"){slug, videoOffsetSeconds}c3:clip(slug:"PuzzledColdbloodedGarbage4Head-XClB8beXWPeSXFf9"){slug, videoOffsetSeconds}c4:clip(slug:"CarelessObliqueApeSquadGoals-KR8NUysH7GJ0nV-u"){slug, videoOffsetSeconds} }\'},\'params\': None, \'headers\': {\'Client-Id\':\'kimne78kx3ncx6brgo4mv6wki5h1ko\'}, \'cookies\':None, \'files\': None, \'auth\': None,\'timeout\': None, \'allow_redirects\':None, \'proxies\': None, \'hooks\':None, \'stream\': None, \'verify\':None, \'cert\': None, \'fu\':furl(\'https://gql.twitch.tv/gql\'), \'domain\':\'gql.twitch.tv\', \'endpoints\': {\'api.ipify.org\':[\'f2zfzh3tq3.execute-api.eu-west-1.amazonaws.com\',\'jt27jndj26.execute-api.eu-west-2.amazonaws.com\',\'iwi6olde13.execute-api.eu-central-1.amazonaws.com\',\'5aan9q8hkb.execute-api.eu-north-1.amazonaws.com\',\'2lv7s6djz2.execute-api.eu-west-3.amazonaws.com\',\'hwxvmq4d0k.execute-api.ca-central-1.amazonaws.com\',\'dtob7epyn2.execute-api.us-east-2.amazonaws.com\',\'7ogy03096l.execute-api.us-east-1.amazonaws.com\',\'bw4siipcr8.execute-api.us-west-1.amazonaws.com\',\'0f3l66pta8.execute-api.us-west-2.amazonaws.com\']}}'], ['gql(query)', '{\'query\': \' query {c0:clip(slug:"AbrasiveGeniusClintTheRinger-eEswgwJ6b0pSfSEC"){slug, videoOffsetSeconds}c1:clip(slug:"HappyArborealClipsmomBlargNaut-Qd9yKSJdoj6ghj1V"){slug, videoOffsetSeconds}c2:clip(slug:"SpineyMuddyPuddingYouDontSay-qYSaw1qvv4bbn_D5"){slug, videoOffsetSeconds}c3:clip(slug:"PuzzledColdbloodedGarbage4Head-XClB8beXWPeSXFf9"){slug, videoOffsetSeconds}c4:clip(slug:"CarelessObliqueApeSquadGoals-KR8NUysH7GJ0nV-u"){slug, videoOffsetSeconds} }\'}'], ['get_clip_offsets(clip_ids)', '{\'clip_ids\': [\'AbrasiveGeniusClintTheRinger-eEswgwJ6b0pSfSEC\', \'HappyArborealClipsmomBlargNaut-Qd9yKSJdoj6ghj1V\', \'SpineyMuddyPuddingYouDontSay-qYSaw1qvv4bbn_D5\', \'PuzzledColdbloodedGarbage4Head-XClB8beXWPeSXFf9\', \'CarelessObliqueApeSquadGoals-KR8NUysH7GJ0nV-u\'], \'slugs\':[\'AbrasiveGeniusClintTheRinger-eEswgwJ6b0pSfSEC\',\'HappyArborealClipsmomBlargNaut-Qd9yKSJdoj6ghj1V\',\'SpineyMuddyPuddingYouDontSay-qYSaw1qvv4bbn_D5\',\'PuzzledColdbloodedGarbage4Head-XClB8beXWPeSXFf9\',\'CarelessObliqueApeSquadGoals-KR8NUysH7GJ0nV-u\'],\'aliases\': [\'c0\', \'c1\', \'c2\', \'c3\', \'c4\'],\'url\': \'https://gql.twitch.tv/gql\',\'query_slugs\': [\'AbrasiveGeniusClintTheRinger-eEswgwJ6b0pSfSEC\', \'HappyArborealClipsmomBlargNaut-Qd9yKSJdoj6ghj1V\', \'SpineyMuddyPuddingYouDontSay-qYSaw1qvv4bbn_D5\', \'PuzzledColdbloodedGarbage4Head-XClB8beXWPeSXFf9\', \'CarelessObliqueApeSquadGoals-KR8NUysH7GJ0nV-u\'], \'query_aliases\': [\'c0\', \'c1\',\'c2\', \'c3\', \'c4\'], \'query_strings\':[\'c0:clip(slug:"AbrasiveGeniusClintTheRinger-eEswgwJ6b0pSfSEC"){slug, videoOffsetSeconds}\',\'c1:clip(slug:"HappyArborealClipsmomBlargNaut-Qd9yKSJdoj6ghj1V"){slug, videoOffsetSeconds}\',\'c2:clip(slug:"SpineyMuddyPuddingYouDontSay-qYSaw1qvv4bbn_D5"){slug, videoOffsetSeconds}\',\'c3:clip(slug:"PuzzledColdbloodedGarbage4Head-XClB8beXWPeSXFf9"){slug, videoOffsetSeconds}\',\'c4:clip(slug:"CarelessObliqueApeSquadGoals-KR8NUysH7GJ0nV-u"){slug, videoOffsetSeconds}\'],\'query\': \' query {c0:clip(slug:"AbrasiveGeniusClintTheRinger-eEswgwJ6b0pSfSEC"){slug, videoOffsetSeconds}c1:clip(slug:"HappyArborealClipsmomBlargNaut-Qd9yKSJdoj6ghj1V"){slug, videoOffsetSeconds}c2:clip(slug:"SpineyMuddyPuddingYouDontSay-qYSaw1qvv4bbn_D5"){slug, videoOffsetSeconds}c3:clip(slug:"PuzzledColdbloodedGarbage4Head-XClB8beXWPeSXFf9"){slug, videoOffsetSeconds}c4:clip(slug:"CarelessObliqueApeSquadGoals-KR8NUysH7GJ0nV-u"){slug, videoOffsetSeconds} }\',\'i\': 0, \'alias_to_slug\': {\'c0\':\'AbrasiveGeniusClintTheRinger-eEswgwJ6b0pSfSEC\', \'c1\':\'HappyArborealClipsmomBlargNaut-Qd9yKSJdoj6ghj1V\', \'c2\':\'SpineyMuddyPuddingYouDontSay-qYSaw1qvv4bbn_D5\', \'c3\':\'PuzzledColdbloodedGarbage4Head-XClB8beXWPeSXFf9\', \'c4\':\'CarelessObliqueApeSquadGoals-KR8NUysH7GJ0nV-u\'},\'slug_to_offset\': {\'AbrasiveGeniusClintTheRinger-eEswgwJ6b0pSfSEC\': None, \'HappyArborealClipsmomBlargNaut-Qd9yKSJdoj6ghj1V\': None, \'SpineyMuddyPuddingYouDontSay-qYSaw1qvv4bbn_D5\': None, \'PuzzledColdbloodedGarbage4Head-XClB8beXWPeSXFf9\': None, \'CarelessObliqueApeSquadGoals-KR8NUysH7GJ0nV-u\': None}}'], ['duration_runner()', "{'CLIP_COUNT': 5, 'slugs':['AbrasiveGeniusClintTheRinger-eEswgwJ6b0pSfSEC','HappyArborealClipsmomBlargNaut-Qd9yKSJdoj6ghj1V','SpineyMuddyPuddingYouDontSay-qYSaw1qvv4bbn_D5','PuzzledColdbloodedGarbage4Head-XClB8beXWPeSXFf9','CarelessObliqueApeSquadGoals-KR8NUysH7GJ0nV-u']}"], ['<module>', "{'get_dictionaries': []}"], ['handler()', '{}']]
    # table = Table()
    # for name, value in L:
    #     w_name = "\n".join(textwrap.wrap(name, 60))
    #     row = (w_name, color_D_if_big(D))
    #     table.row(row)
    # print(table)
    #
def main():
    return divide_by_zero()

if __name__ == "__main__":
    from .timer import Timer
    from .ic import ic; ic
    timer = Timer()
    with handler():
        main()



# COPY PASTE THIS
# import sys
# sys.path = [ "C:/users/vivek/Documents/Python/" ] + sys.path

# How to Install: 
# 
#   pip install https://github.com/waivek/python_common/archive/master.zip
#    
#    PHP FUNCTIONS:
#       (list)  array_chunk
#       (dict)  array_count_values
#       (table) array_column
#       (dict)  array_extrace
#
# Refactory:
#
#       abbr(number)
#       trunc(string, length)
#       pad_zero(current_value, max_value)
#       pad_space(current_value, max_value)
#       enumerate_count(items)
#       rel2abs(relative_path)
#
from .timer import Timer
from .color import Code
timer = Timer()

import sys
import os.path

RANGE_DEPTH = 8

def simple_repr(self):
    items = ("%s=%r" % (k, v) for k, v in self.__dict__.items())
    return "%s(%s)" % (self.__class__.__name__, ', '.join(items))

def rel2abs(relative_path):
    from .reltools import rel2abs as func
    return func(relative_path)

def print_error_information(error):
    from .error import print_error_information as func
    func(error)

def stub_quiet():
    t = Timer()

# --- START
def print_function(function_name, for_type, frame_index_strings):
    from .color import Code
    for string in frame_index_strings:
        left, right = string.split(" --- ")
        if right == r"C:\Users\vivek\Desktop\Twitch\coffee-vps\code":
            right = Code.GREEN + right
        elif right != r"C:\Users\vivek\Documents\Python":
            right = Code.RED + right
        print(f"{for_type:4s}:{function_name:30s} --- {left:24s} : {right}")
    print()

def frame_index_to_directory(frame_index):
    try:
        frame            = sys._getframe(frame_index)
    except Exception as e:
        return "None --- None"
    saved_frame = frame
    called_path      = frame.f_code.co_filename
    parent_directory = os.path.dirname(os.path.realpath(called_path))
    co_name = frame.f_code.co_name
    return f"{co_name} --- {parent_directory}"

def tplib(tag):

    lines = [ frame_index_to_directory(i) for i in range(RANGE_DEPTH) ]
    print_function(tag + "-> tplib", "comp", lines)
    lines = []
    for i in range(RANGE_DEPTH):
        lines.append(frame_index_to_directory(i))
    print_function(tag + "-> tplib", "for", lines)



def create_partitions_2(size, n):
    starts = list(range(0, size, n))
    ends = starts[1:] + [ size ]
    return list(zip(starts, ends))

def create_partitions_old(max_value, partition_value):
    import math
    number_of_partitions = math.ceil(float(max_value) / partition_value)
    starts = [ i * partition_value for i in range(number_of_partitions) ]
    ends = starts[1:] + [max_value]
    return [(start, end) for start, end in zip(starts, ends)]

def create_partitions(max_value, partition_value):
    indices = list(range(0, max_value, partition_value)) + [ max_value ]
    return list(zip(indices, indices[1:]))


def abbreviate(number):
    number_int = int(number)
    number_string = str(number_int)
    number_length = len(str(number_int))
    if number_int < 100:
        return number_string
    first_digit, second_digit, third_digit = number_string[0:3]
    abbreviation_dict = {
         1:  "",  2:  "",  3:  "",
         4: "k",  5: "k",  6: "k",
         7: "m",  8: "m",  9: "m",
        10: "b", 11: "b", 12: "b",
        13: "t", 14: "t", 15: "t"
    }
    letter = abbreviation_dict[number_length]
    if number_length % 3 == 0:
        # first three digits
        abbr_string = f"{first_digit}{second_digit}{third_digit}{letter}"
    if number_length % 3 == 2:
        # first two digits, period, next digit
        abbr_string = f"{first_digit}{second_digit}.{third_digit}{letter}"
    if number_length % 3 == 1:
        # first digit, period, next two digits
        abbr_string = f"{first_digit}.{second_digit}{third_digit}{letter}"

    return abbr_string

def smart_pad(current_value, max_value, fillchar='0'):
    number_of_digits = len(str(max_value))
    return str(current_value).rjust(number_of_digits, fillchar)

def enumerate_count(items):
    item_count = len(items)
    count_strings = []
    for i in range(item_count):
        count_str = smart_pad(i+1, item_count)
        count_strings.append(f"[{count_str}/{item_count}]")
    return zip(count_strings, items)

def print_object(obj, hidden=False):
    if not hidden:
        keys = [ key for key in dir(obj) if not key.startswith("__") ]
    else:
        keys = dir(obj)
    D = { key : obj.__getattribute__(key) for key in keys }
    print_dict(D)


# print_table {{{
# Assumption: Columns have consistent type
def print_table(dictionaries):
    if len(dictionaries) == 0:
        return
    for D in dictionaries:
        for key in D.keys():
            value = D[key]
            if type(value) != str and type(value) != int and type(value) != float:
                D[key] = str(value)
    # keys = ["basename", "file_read_time", "json_load_time", "total_time"]
    keys = list(dictionaries[0].keys())
    # keys = ["name", "cost", "count"]
    max_length_D = {key:len(key) for key in keys}
    types_D = { key: type(dictionaries[0][key]) for key in keys }
    left_gutter_length = 2
    vertical_padding = 1
    left_gutter_string = " " * left_gutter_length
    column_gap = 2
    separator = " "  * column_gap
    for key in keys:
        max_length = max_length_D[key]
        for D in dictionaries:
            value = D[key]
            if type(value) == str:
                length = len(value)
            if type(value) == int:
                length = len(str(value))
            if type(value) == float:
                length = len("%.2f" % value)
            if max_length < length:
                max_length = length
        max_length_D[key] = max_length

    for i in range(vertical_padding):
        print()
    column_label_items = []
    for key in keys:
        if types_D[key] == str:
            rep = key.ljust(max_length_D[key])
        if types_D[key] == int:
            rep = key.rjust(max_length_D[key])
        if types_D[key] == float:
            rep = key.rjust(max_length_D[key])
        column_label_items.append(rep)
    line = left_gutter_string + separator.join(column_label_items)
    print(line)


    for D in dictionaries:
        line_items = []
        for key in keys:
            max_length = max_length_D[key]
            value = D[key]
            if type(value) == str:
                rep = value.ljust(max_length)
            if type(value) == int:
                rep = str(value).rjust(max_length)
            if type(value) == float:
                rep = ("%.2f" % value).rjust(max_length)
            line_items.append(rep)
        line = left_gutter_string + separator.join(line_items)
        print(line)
    for i in range(vertical_padding):
        print()
        # }}}

def make_pretty(json_filepath):
    import json
    with open(json_filepath, "r") as f:
        dictionaries = json.load(f)
    filepath, extension = os.path.splitext(json_filepath)
    pretty_json_filepath = filepath + f"-pretty{extension}"
    with open(pretty_json_filepath, "w") as f:
        json.dump(dictionaries, f, indent=4)

    message = f"{json_filepath} -> {pretty_json_filepath}"
    print(message)

def to_json(obj, filepath):
    import json
    with open(filepath, "w") as fp:
        json.dump(obj, fp, indent=4)

def head_file(file_contents, preview_lines_count=5):
    lines = file_contents.split("\n")
    preview_lines = lines[0:preview_lines_count]
    return "\n".join(preview_lines)

def head_list(L, preview_lines_count=5):
    preview_L = L[0:preview_lines_count]
    return "\n".join(preview_L)

def head(inp, preview_lines_count=5):
    if type(inp) == list:
        return head_list(inp, preview_lines_count)
    else:
        return head_file(inp, preview_lines_count)

def print_dict(D):
    from columnar import columnar
    if type(D) != type({}):
        argument_type = str(type(D))
        print_red_line("Please pass a dictionary. Invalid type {arg_type} with str representation {str_rep}".format(arg_type=argument_type, str_rep=str(D)))
        return
    data = []
    for key, value in D.items():
        trunc_val = truncate(str(value), 80)
        trunc_val = str(value)
        data.append([key, trunc_val])
    headers = [ "key", "value" ]
    table = columnar(data, headers, no_borders=True)
    print(table)
            

def file_to_string(file_path):
    if not os.path.exists(file_path):
        return ""
    with open(file_path, "rb") as f:
        file_contents = f.read().decode("utf-8")
    return file_contents

def read_json_cache(cache_path):
    import json
    cache_contents = file_to_string(cache_path)
    if cache_contents == "":
        cache_contents = "{}"
    cache_D = json.loads(cache_contents)
    return cache_D

def update_json_cache(cache_path, new_cache_D):
    import json
    cache_contents = file_to_string(cache_path)
    if cache_contents == "":
        cache_contents = "{}"
    old_cache_D = json.loads(cache_contents)
    if old_cache_D == new_cache_D:
        return

    new_cache_contents = json.dumps(new_cache_D)
    with open(cache_path, "wb") as f:
        f.write(new_cache_contents.encode("utf-8"))
        print("[WRITE] {filename}".format(filename=f.name))


def make_string_green(string):
    # from .colorama import init, Fore
    # init(convert=True)
    # return "{color_code}{string}{reset_code}".format(color_code=Fore.GREEN, string=string, reset_code=Fore.RESET)
    return Code.GREEN + string
def make_string_red(string):
    # from .colorama import init, Fore
    # init(convert=True)
    # return "{color_code}{string}{reset_code}".format(color_code=Fore.RED, string=string, reset_code=Fore.RESET)
    return Code.RED + string
def print_red_line(string):
    print(make_string_red(string))
def print_green_line(string):
    print(make_string_green(string))

# for copy-pasting into libraries
def truncate_2(s, L):
    # M: middle_part, S: slice_size, s: string, T: trunc_string, R: right_start, L: length
    if len(s) <= L:
        return s
    M = "..."
    S = int((L-len(M))/2)
    R = S + int(L % 2 == 0)
    T = s[0:S]+M+s[-R:]
    return T if len(T) == L else s

def truncate(string, length):
    if len(string) <= length:
        return string

    middle_string = "..."
    slice_size = int((length-len(middle_string))/2)
    # 80, 38, 38, 3 SUM = 79/80 ONE SPACE AVAILABLE
    # 81, 39, 39, 3 SUM = 81/81 ZERO SPACES AVAILABLE
    even_string_length = length % 2 == 0
    left_end = slice_size
    if even_string_length:
        right_start = slice_size+1
    else:
        right_start = slice_size
    left_slice = string[0:left_end]
    right_slice = string[-right_start:]
    trunc_string = "{left_slice}{middle_string}{right_slice}".format(left_slice=left_slice, middle_string=middle_string, right_slice=right_slice)
    if len(trunc_string) != length:
        print_red_line("ERROR: Could not return a truncated string")
        return "ERROR"
    return trunc_string

def read_stdin():
    # sys.stdout.reconfigure(encoding='utf-8')
    input_stream = sys.stdin
    file_contents = input_stream.read()
    return file_contents
def read_pipe_or_file():
    import argparse
    stream_is_interactive = sys.stdin.isatty()
    if stream_is_interactive:
        parser = argparse.ArgumentParser(description="Removes attributes from a HTML file")
        parser.add_argument("input_filename")
        args = parser.parse_args()
        with open(args.input_filename, "rb") as f:
            file_contents = f.read().decode("utf-8")
    else:
        file_contents = read_stdin()
    return file_contents

class Timestamp():
    def __init__(self, tstamp):
        import re
        if type(tstamp) == str and re.match(r'^[0-9.]+$', tstamp):
            tstamp = int(tstamp)
        if type(tstamp) == str:
            m = re.match(r'\d?\dhh?\d?\dmm?\d?\ds?', tstamp)
            if m:
                # tstamp is in 01h02m03s format
                tstamp = re.sub(r'[hm]+', ":", tstamp).replace("s", "")
            a, b, c = tstamp.split(":")
            self.hh = int(a)
            self.mm = int(b)
            self.ss = int(float(c))
            self.timestamp = tstamp
            self.seconds = self.hh * 3600 + self.mm * 60 + self.ss
        elif type(tstamp) == float or type(tstamp) == int:
            seconds_placeholder = int(tstamp)
            self.seconds = seconds_placeholder
            self.hh = int(seconds_placeholder / 3600)
            seconds_placeholder = seconds_placeholder % 3600
            self.mm = int(seconds_placeholder / 60)
            seconds_placeholder = seconds_placeholder % 60
            self.ss = seconds_placeholder 
        self.timestamp = "{hours:02}:{minutes:02}:{seconds:02}".format(hours=self.hh, minutes=self.mm, seconds=self.ss)
        self.hh_mm_ss = "{hours:02}h{minutes:02}m{seconds:02}s".format(hours=self.hh, minutes=self.mm, seconds=self.ss)
    def __repr__(self):
        if self.seconds == 1:
            return "00:00:01: 1 second" 
        return "%s: %d seconds" % (self.timestamp, self.seconds)
    def __add__(a, b):
        return Timestamp(a.seconds + b.seconds)
    def __sub__(a, b):
        return Timestamp(abs(a.seconds - b.seconds))
    def days(self):
        d = int(self.hh % 24)
        hh = self.hh / 24
        return "%d days %02d:%02d:%02d" % (d, hh, self.mm, self.ss)


class Date:
    def __init__(self, string_or_datetime_or_epoch):
        import dateutil.parser
        from datetime import datetime
        from datetime import timezone
        from datetime import timedelta
        if isinstance(string_or_datetime_or_epoch, datetime):
            dt_str = string_or_datetime_or_epoch.isoformat()
        elif isinstance(string_or_datetime_or_epoch, int) or isinstance(string_or_datetime_or_epoch, float):
            dt_str = datetime.fromtimestamp(string_or_datetime_or_epoch).isoformat()
        else:
            dt_str = string_or_datetime_or_epoch
        # INDIAN_TIMEZONE  = dateutil.tz.gettz("Asia/Kolkata")
        INDIAN_TIMEZONE = timezone(timedelta(hours=5, minutes=30), 'IST')

        self.dt = dateutil.parser.parse(dt_str).astimezone(INDIAN_TIMEZONE).replace(microsecond=0)
        self.string = self.dt.isoformat()
        self.epoch = int(self.dt.timestamp())

    def timeago(self):
        from datetime import timezone
        import timeago
        from datetime import datetime
        now = datetime.now(timezone.utc)
        return timeago.format(self.dt, now)

    def now():
        from datetime import datetime
        return Date(datetime.now())

    def pretty(self):
        from datetime import timezone
        print_dict({
            "ist": self.string,
            "utc": self.dt.astimezone(timezone.utc).isoformat()[:-6],
            "epoch": self.epoch,
            "timeago": self.timeago()
        })

    # `=`, `+`, `-`, `<`, `>`, str, repr {{{
    def __eq__(self, rhs_date):
        return self.dt == rhs_date.dt
    def __add__(self, seconds_rhs):
        seconds = int(seconds_rhs)
        from datetime import timedelta
        return Date(self.dt + timedelta(seconds=seconds))
    def __sub__(self, rhs):
        # Date - int      = Date (earlier date)
        # Date - Date     = int  (total seconds)
        # Date - datetime = int  (total seconds)

        from datetime import timedelta
        from datetime import datetime
        if type(rhs) == int:
            difference_datetime = self.dt - timedelta(seconds=rhs)
            return Date(difference_datetime)
        elif type(rhs) == Date:
            difference_timedelta = self.dt - rhs.dt
            return difference_timedelta.total_seconds()
        elif type(rhs) == datetime:
            difference_timedelta = self.dt - rhs
            return difference_timedelta.total_seconds()

    def __lt__(self, other):
        return self.dt < other.dt
    def __gt_(self, other):
        return self.dt > other.dt
    def __str__(self):
        timeago_str = self.timeago()
        return "{formatted_datetime} ({timeago_string})".format(
                formatted_datetime=self.dt.strftime("%I:%M %p, %b %d"),
                timeago_string=timeago_str)
    def __repr__(self):
        repr_string = repr(self.string)
        return f"Date({repr_string})"
    def __format__(self, string):
        return self.dt.strftime(string)
    # }}}

if __name__ == "__main__":
    from .ic import ic
    error_table = [{'slug': 'ThankfulDeadMinkJonCarnage-A1wuckc0t442Zdzx', 'views': 384, 'user_id': '160504245', 'title': 'lacari playing gwen', 'thumbnail_url': 'https://clips-media-assets2.twitch.tv/AT-cm%7CtyYhdQNcLnJHNLeONmXc8Q-preview-480x272.jpg'}]
    ic(error_table)
    pass




# Asynchronous Getting
# 0. Get JSON / READ / TEXT / RESPONSE (sync)
# 1. Toggleable Caching - render_chat.py:get_images_cached(urls)
# 2. Configurable Connection Limit - render_chat.py:get_json_batch(urls, limit=100)
# 3. Toggleable AsyncInfo - Log status url reason counts
# 4. Handle status codes 404 and non-200
# 5. Configureable url2header Function
# 6. POST with payload

from .timer import Timer   # Single Use
timer = Timer()

from .db import db_init
from .color import Code

# import aiohttp
# import asyncio
import json

aiohttp = None
asyncio = None
connection = None

class Config:
    pass

class Response:
    def __init__(self, response, obj):
        self.url = str(response.url)
        self.status = response.status
        self.ok = response.ok
        self.reason = response.reason
        self.headers = {row[0]: row[1] for row in response.headers.items()}
        self.obj = obj
        self.total_bytes = response.content.total_bytes
    def manual_init():
        url = None
        status = 200
        ok = True
        reason = "OK"
        headers = {}
    def __repr__(self):
        return f"Response <{self.status}>"


class AsyncInfo:
    def __init__(self, lock):
        self.lock = lock
        self.results = []
        self.total = 0
        self.count_pass = 0
        self.count_fail = 0
        self.code_counter = {}

    def add(self, response):
        url = response.url
        reason = response.reason
        status = response.status

        self.results.append({ 'url': url, 'status': status, 'reason': reason })
        self.total = self.total + 1 
        if status == 200:
            self.count_pass = self.count_pass + 1
        else:
            self.count_fail = self.count_fail + 1
        self.code_counter[status] = self.code_counter.get(status, 0) + 1

    def print(self):
        code_strings = []
        for code in sorted(self.code_counter.keys()):
            count = self.code_counter[code]
            color = Code.GREEN if code < 400 else Code.RED
            code_strings.append(color + f"{code}: {count}")
        code_string = "{ " + ", ".join(code_strings) + " }"
        print(f"Total: {self.total} Pass: {self.count_pass} Fail: {self.count_fail} Codes: {code_string}", end="\r")

def configure_asyncio():
    global asyncio
    global aiohttp
    import asyncio
    import aiohttp

    if Config.log:
        global async_info
        async_info = AsyncInfo(asyncio.Lock())

    import platform
    if platform.system() == 'Windows':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def is_binary(content_type):
    mime_type, subtype = content_type.split("/")

    if mime_type == "text":
        return False
    if mime_type != "application":
        return True
    return subtype not in ["json", "ld+json", "x-httpd-php", "x-sh", "x-csh", "xhtml+xml", "xml"]

# read text json content_type
async def async_response_to_sync_response(response):
    # https://stackoverflow.com/questions/57565577/how-to-return-aiohttp-like-python-requests
    # https://docs.aiohttp.org/en/stable/client_reference.html#response-object
    if response.content_type == 'application/json':
        obj = await response.json()
    elif is_binary(response.content_type):
        obj = await response.read()
    else:
        obj = await response.text()
    return Response(response, obj)

async def log_response(response):
    if Config.log:
        async with async_info.lock:
            async_info.add(response)
            async_info.print()

async def get_async(url, session):
    headers = Config.url2header(url)
    cookies = Config.url2cookie(url)
    # breakpoint()
    response = await session.get(url, headers=headers, cookies=cookies)
    async with response:
        sync_response = await async_response_to_sync_response(response)
    await log_response(response)

    return sync_response

async def batch_get_async(urls):
    async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(limit=Config.limit), trust_env=True) as session:
        cors = [ get_async(url, session) for url in urls ]
        responses = await asyncio.gather(*cors)
    if Config.log and len(urls) > 0:
        print()
    return responses

def get_nocache(urls):
    configure_asyncio()
    # [ obj ]
    responses = asyncio.run(batch_get_async(urls))
    return responses

# -- START: CACHE --
def to_temp_table(cursor, items, table_string):
    table_name = table_string.split(" ")[0]
    cursor.execute(f"CREATE TEMP TABLE {table_string};")
    cursor.executemany(f"INSERT INTO {table_name} VALUES (?);", [ (item,) for item in items ])

def get_cached_urls(urls):
    cursor = connection.cursor()
    to_temp_table(cursor, urls, "urls (url TEXT UNIQUE)")
    uncached_urls = [ url for url, in cursor.execute("SELECT url FROM urls WHERE url IN (SELECT url FROM cache);").fetchall() ]
    cursor.execute("DROP TABLE urls;")
    return uncached_urls


def get_uncached_urls(urls):
    cursor = connection.cursor()
    to_temp_table(cursor, urls, "urls (url TEXT UNIQUE)")
    uncached_urls = [ url for url, in cursor.execute("SELECT url FROM urls WHERE url NOT IN (SELECT url FROM cache);").fetchall() ]
    cursor.execute("DROP TABLE urls;")
    return uncached_urls

def object_to_type(obj):
    if type(obj) is str:
        return 'text'
    elif type(obj) is bytes:
        return 'binary'
    else:
        return 'json'

def cache(urls, objects):
    cursor = connection.cursor()
    tuples = []
    for url, obj in zip(urls, objects):
        obj_type = object_to_type(obj)
        obj = json.dumps(obj) if obj_type == 'json' else obj
        T = (url, obj_type, obj)
        tuples.append(T)

    cursor.executemany("INSERT INTO cache (url, type, response) VALUES (?, ?, ?);", tuples)
    connection.commit()

def cache_get(urls):
    cursor = connection.cursor()
    to_temp_table(cursor, urls, "urls (url TEXT UNIQUE)")
    url_to_object_D = { url: response for url, response in cursor.execute("SELECT url, response FROM cache WHERE url IN (SELECT url FROM urls);") }
    url_to_type_D = { url: type for url, type in cursor.execute("SELECT url, type FROM cache WHERE url IN (SELECT url FROM urls);") }
    objects = []
    for url in urls:
        object_type = url_to_type_D[url]
        obj = url_to_object_D[url]
        if object_type == 'json':
            obj = json.loads(obj)
        objects.append(obj)
    cursor.execute("DROP TABLE urls;")
    return objects

def get_yescache(urls):
    # -> [ obj | Response ]
    cursor = connection.cursor()
    if DEBUG:
        cursor.execute("DROP TABLE IF EXISTS cache;")
    cursor.execute("CREATE TABLE IF NOT EXISTS cache (url TEXT UNIQUE, type TEXT, response);")

    unique_urls = list(set(urls))
    uncached_urls = get_uncached_urls(unique_urls)
    responses = get_nocache(uncached_urls) if len(uncached_urls) > 0 else []
    pass_objects = [ response.obj for response in responses if response.ok ]
    cache(uncached_urls, pass_objects)

    cached_urls = get_cached_urls(unique_urls)
    cached_objects = cache_get(cached_urls)
    D1 = { url: obj for url, obj in zip(cached_urls, cached_objects) }
    D2 = { url: response.obj if response.ok else response for url, response in zip(uncached_urls, responses) }
    url_to_object_D = D1 | D2 # Dictionary Union
    objects = [ url_to_object_D[url] for url in urls ]
    return objects

# -- END: CACHE --

def aget(urls, cache=True, log=True, limit=100, url2header=lambda url: {}, url2cookie=lambda url: {}):
    # START: Configuration

    global DEBUG
    DEBUG = False
    Config.cache = cache
    Config.log = log
    Config.url2header = url2header
    Config.url2cookie = url2cookie
    Config.limit = limit
    if Config.cache:
        global connection
        connection = db_init("aget_cache/cache.db")[1]
    # END: Configuration
    if DEBUG:
        print(Code.RED + "Debug: True")
    if cache:
        objects = get_yescache(urls)
        return objects
    else:
        responses = get_nocache(urls)
        objects = [ response.obj for response in responses ]
        return objects

def run_async_get():
    json_url  = "https://api.betterttv.net/3/cached/emotes/global"
    image_url = "https://cdn.betterttv.net/emote/56e9f494fff3cc5c35e5287e/3x"
    html_url  = "https://example.com/"
    text_url = "https://www.w3.org/TR/PNG/iso_8859-1.txt"
    dupe_url = "https://www.google.com"
    not_found_url = "https://cdn.betterttv.net/emote/404/3x"

    urls = [ json_url, image_url, html_url, text_url, not_found_url, dupe_url, dupe_url ]
    objects = aget(urls, cache=True, log=True)
    ic(objects)
    return objects

if __name__ == "__main__":
    from .ic import ic
    ic
    from .error import handler
    with handler():
        run_async_get()




from time import time
class Timer():

    def __init__(self, no_print=False):
        self.start_time = None
        self.inc_start = None
        self.timer_D = {}
        self.no_print = no_print
        self.sum_D = {}

    def start(self, message=None):
        self.start_time = time()
        if message:
            self.timer_D[message] = time()

    def get(self, message):
        if self.timer_D.get(message) == None and self.start_time == None:
            print("ERROR: Timer configured incorrectly")

        time_taken = time() - self.timer_D.get(message, self.start_time)
        if message in self.timer_D:
            value = self.timer_D[message]
            del self.timer_D[message]
        return time_taken

    def start_inc(self):
        self.inc_start = time()
    def print_inc(self, message):
        if self.no_print:
            return
        time_taken = time() - self.inc_start
        print(f"{message:20s}: {time_taken:.2f} seconds")
        self.inc_start = time()



    def start_multi(self):
        self.multi_start_time = time()

    def print_multi(self, message):
        if self.no_print:
            return
        time_taken = time() - self.multi_start_time
        print(f"{message:20s}: {time_taken:.2f} seconds")

    def start_sum(self, message):
        self.timer_D[message] = time()

    def add_sum(self, message):
        time_taken = time() - self.timer_D[message]
        current_sum = self.sum_D.get(message, 0)
        self.sum_D[message] = current_sum + time_taken

    def print_sum(self, message):
        time_taken = self.sum_D[message]
        print(f"{message:20s}: {time_taken:.2f} seconds")
        del self.sum_D[message]

    def print(self, message):
        if self.no_print:
            return
        if self.timer_D.get(message) == None and self.start_time == None:
            print("ERROR: Timer configured incorrectly")
        time_taken = time() - self.timer_D.get(message, self.start_time)
        print(f"{message:20s}: {time_taken:.2f} seconds")
        if self.timer_D.get(message):
            del self.timer_D[message]
        else:
            self.start_time = None


# <space>T normal ToggleTimeFile
# <space>t visual TimeSelection
# <space>t normal on_timer_line ? DeleteTimer : TimeLine
def cases():
    from time import sleep
    timer = Timer()
    # If there are multiple lines printed, only 1 can exist at a time,
    # otherwise somewhere we'll require two arguments

    # Variant 1
    timer.start("sleep")
    sleep(1)
    timer.print("sleep")
    # --- 

    # Variant 2
    timer.start()
    sleep(1)
    timer.print("quick-sleep")
    # --- 

    # Variant 3
    timer.start_loop()
    for i in range(5):
        sleep(0.1)
        timer.add_loop()
    timer.print_loop("Loop Time")
    # --- 

    # Variant 4 (most used pattern, we have consecutive parts in a function that we want to arbitrarily break down and profile)
    import datetime
    dt = datetime
    timer.print_inc("import datetime")
    # --- 

    # Variant 5 (should be main mode of profiling, so that other prints don’t interfere and if you have an error, you don’t care about performances)
    # Don’t interfere with other table that’s being printed
    timer.start_agg() # ID-less, only 1 can exist at a time, otherwise log would require two arguments
    sleep(0.1)
    print("Time slept for: 100 ms")
    timer.log_agg("First Sleep Time")
    sleep(0.2)
    print("Time slept for: 200 ms")
    timer.log_agg("Second Sleep Time")
    timer.print_agg()
    # --- 

